"""Test neurots.extract_input code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
import os

import neurom
import numpy as np
import pytest
import tmd
from neurom import load_morphologies
from neurom import stats
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_equal

from neurots import NeuroTSError
from neurots import extract_input
from neurots import validator

_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "test_data")
POP_PATH = os.path.join(_PATH, "bio/")
NEU_PATH = os.path.join(_PATH, "diam_simple.swc")


@pytest.fixture
def POPUL():
    return load_morphologies(POP_PATH)


@pytest.fixture
def NEU():
    return load_morphologies(NEU_PATH)


def test_num_trees(POPUL):
    target_numBAS = {
        "num_trees": {"data": {"bins": [4, 5, 6, 7, 8, 9], "weights": [1, 0, 0, 0, 0, 1]}}
    }
    target_numAX = {"num_trees": {"data": {"bins": [1], "weights": [2]}}}

    numBAS = extract_input.from_neurom.number_neurites(POPUL)
    numAX = extract_input.from_neurom.number_neurites(POPUL, neurite_type=neurom.AXON)
    assert_equal(numBAS, target_numBAS)
    assert_equal(numAX, target_numAX)


def test_trunk_distr(POPUL):
    bins_BAS = [
        0.19391773616376634,
        0.4880704446023673,
        0.7822231530409682,
        1.076375861479569,
        1.3705285699181702,
        1.6646812783567713,
        1.9588339867953721,
        2.2529866952339734,
        2.547139403672574,
        2.841292112111175,
    ]

    absolute_elevation_deviation_BAS = {
        "data": {
            "weights": [2, 0, 0, 3, 1, 2, 1, 1, 1, 2],
        }
    }
    bins_absolute_ele_dev_BAS = [
        -0.7718245274301169,
        -0.6464835753472811,
        -0.5211426232644452,
        -0.39580167118160936,
        -0.27046071909877345,
        -0.1451197670159376,
        -0.019778814933101796,
        0.10556213714973406,
        0.23090308923256997,
        0.35624404131540577,
    ]
    bins_absolute_ele_dev_APIC = [1.03738723]

    target_trunkBAS = {
        "trunk": {
            "azimuth": {"uniform": {"max": 0.0, "min": np.pi}},
            "orientation_deviation": {"data": {"weights": [4, 3, 1, 2, 0, 1, 0, 0, 0, 2]}},
            "absolute_elevation_deviation": absolute_elevation_deviation_BAS,
        }
    }
    target_trunkAPIC = {
        "trunk": {
            "azimuth": {"uniform": {"max": 0.0, "min": np.pi}},
            "orientation_deviation": {"data": {"bins": [0.0], "weights": [2]}},
            "absolute_elevation_deviation": {"data": {"weights": [2]}},
        }
    }

    trunkAP = extract_input.from_neurom.trunk_neurite(
        POPUL, neurite_type=neurom.APICAL_DENDRITE, bins=1
    )
    trunkBAS = extract_input.from_neurom.trunk_neurite(POPUL, bins=10)

    assert_array_almost_equal(trunkBAS["trunk"]["orientation_deviation"]["data"]["bins"], bins_BAS)
    assert_array_almost_equal(
        trunkBAS["trunk"]["absolute_elevation_deviation"]["data"]["bins"],
        bins_absolute_ele_dev_BAS,
    )
    assert_array_almost_equal(
        trunkAP["trunk"]["absolute_elevation_deviation"]["data"]["bins"],
        bins_absolute_ele_dev_APIC,
    )
    del trunkBAS["trunk"]["orientation_deviation"]["data"]["bins"]
    del trunkBAS["trunk"]["absolute_elevation_deviation"]["data"]["bins"]
    del trunkAP["trunk"]["absolute_elevation_deviation"]["data"]["bins"]

    assert_equal(trunkBAS, target_trunkBAS)
    assert_equal(trunkAP, target_trunkAPIC)


def test_diameter_extract(POPUL, NEU):
    res = extract_input.from_diameter.model(NEU)
    assert_equal(set(res.keys()), {"basal_dendrite"})
    expected = {
        "Rall_ratio": 1.5,
        "siblings_ratio": 1.0,
        "taper": [0.24, 0.1],
        "term": [2.0, 2.0],
        "trunk": [3.9],
        "trunk_taper": [0.30],
    }

    assert_equal(set(res["basal_dendrite"]), set(expected))
    for key, value in expected.items():
        assert_array_almost_equal(res["basal_dendrite"][key], value)

    with pytest.raises(NeuroTSError):
        extract_input.from_diameter.model(load_morphologies(os.path.join(_PATH, "simple.swc")))

    # Test on Population
    res = extract_input.from_diameter.model(POPUL)
    assert_equal(set(res.keys()), {"axon", "basal_dendrite", "apical_dendrite"})
    expected = {
        "basal_dendrite": {
            "Rall_ratio": 1.5,
            "siblings_ratio": 1.0,
            "taper": [
                0.003361,
                0.009487,
                0.009931,
                0.016477,
                0.023878,
                0.024852,
                0.027809,
                0.027975,
            ],
            "term": [0.3] * 8,
            "trunk": [0.6, 0.6, 0.72, 0.84, 1.2, 1.5, 1.8, 2.4],
            "trunk_taper": [
                0,
                3.036411e-02,
                3.053287e-02,
                5.059035e-02,
                1.168936e-01,
                1.172027e-01,
                0.15,
                2.121002e-01,
            ],
        },
        "apical_dendrite": {
            "Rall_ratio": 1.5,
            "siblings_ratio": 1.0,
            "taper": [
                0.010331,
                0.02135,
                0.02264,
                0.033914,
                0.035313,
                0.041116,
                0.055751,
                0.056211,
            ],
            "term": [0.3] * 8,
            "trunk": [1.57, 7.51],
            "trunk_taper": [0.05324615, 0.65223652],
        },
        "axon": {
            "Rall_ratio": 1.5,
            "siblings_ratio": 1.0,
            "taper": [
                0.04079,
                0.055286,
                0.092382,
                0.099524,
                0.11986,
                0.140346,
                0.214172,
                0.407058,
            ],
            "term": [0.12] * 8,
            "trunk": [2.1, 3.0],
            "trunk_taper": [0.0435508, 0.0717109],
        },
    }

    for neurite_type in ["basal_dendrite", "apical_dendrite", "axon"]:
        for key in expected[neurite_type]:
            try:
                assert_equal(res[neurite_type].keys(), expected[neurite_type].keys())
                if key in ["taper", "term", "trunk", "trunk_taper"]:
                    tested = sorted(res[neurite_type][key])[:8]
                else:
                    tested = res[neurite_type][key]
                assert_array_almost_equal(tested, expected[neurite_type][key])
            except AssertionError as err:
                raise AssertionError(f"Failed for res[{neurite_type}][{key}]") from err


class TestDistributions:
    """Test distributions."""

    @pytest.fixture
    def filename(self):
        return os.path.join(_PATH, "bio/")

    def test_radial_distances(self, filename):
        distr = extract_input.distributions(filename, feature="radial_distances")
        assert_equal(
            set(distr.keys()), {"soma", "basal_dendrite", "apical_dendrite", "axon", "diameter"}
        )
        distr_legacy = extract_input.distributions(
            filename, feature="radial_distances", neurite_types=["basal", "apical", "axon"]
        )
        assert_equal(
            set(distr_legacy.keys()),
            {"soma", "basal_dendrite", "apical_dendrite", "axon", "diameter"},
        )
        assert_equal(
            distr["basal_dendrite"]["num_trees"],
            {"data": {"bins": [4, 5, 6, 7, 8, 9], "weights": [1, 0, 0, 0, 0, 1]}},
        )
        assert_equal(distr["basal_dendrite"]["filtration_metric"], "radial_distances")

        # Check that the returned distributions are valid according to the JSON schema
        validator.validate_neuron_distribs(distr)

    def test_path_distances(self, filename):
        distr = extract_input.distributions(filename, feature="path_distances")
        assert_equal(distr["basal_dendrite"]["filtration_metric"], "path_distances")
        validator.validate_neuron_distribs(distr)

    def test_trunk_length(self, filename):
        distr = extract_input.distributions(filename, feature="trunk_length")
        assert "persistence_diagram" not in distr["basal_dendrite"]
        assert "filtration_metric" not in distr["basal_dendrite"]
        validator.validate_neuron_distribs(distr)

    def test_different_features(self, filename):
        # Test with different features for each neurite type
        distr = extract_input.distributions(
            filename,
            feature={
                "apical_dendrite": "radial_distances",
                "basal_dendrite": "trunk_length",
                "axon": "path_distances",
            },
        )
        assert "filtration_metric" not in distr["basal_dendrite"]
        assert distr["apical_dendrite"]["filtration_metric"] == "radial_distances"
        assert distr["axon"]["filtration_metric"] == "path_distances"
        assert "persistence_diagram" not in distr["basal_dendrite"]
        assert "persistence_diagram" in distr["apical_dendrite"]
        assert "persistence_diagram" in distr["axon"]
        validator.validate_neuron_distribs(distr)

    def test_diameter_model_none(self, filename):
        distr = extract_input.distributions(
            filename,
            feature="radial_distances",
            diameter_model=None,
        )
        assert_equal(
            set(distr.keys()), {"soma", "basal_dendrite", "apical_dendrite", "axon", "diameter"}
        )
        validator.validate_neuron_distribs(distr)

    def test_diameter_model_M5(self, filename):
        distr_M5 = extract_input.distributions(
            filename, feature="radial_distances", diameter_model="M5"
        )
        assert_equal(
            set(distr_M5.keys()), {"soma", "basal_dendrite", "apical_dendrite", "axon", "diameter"}
        )
        validator.validate_neuron_distribs(distr_M5)

    def test_external_diameter_model(self, filename):
        def diam_method(pop):
            return extract_input.from_diameter.model(pop)

        distr_external = extract_input.distributions(
            filename, feature="radial_distances", diameter_model=diam_method
        )
        assert_equal(
            set(distr_external.keys()),
            {"soma", "basal_dendrite", "apical_dendrite", "axon", "diameter"},
        )
        validator.validate_neuron_distribs(distr_external)

        distr_external_input = extract_input.distributions(
            filename,
            feature="radial_distances",
            diameter_model=diam_method,
            diameter_input_morph=filename,
        )
        assert distr_external == distr_external_input


def test_transform_distr():
    ss = stats.fit([1, 2], distribution="norm")
    tss = extract_input.from_neurom.transform_distr(ss)
    assert_equal(tss, {"norm": {"mean": 1.5, "std": 0.5}})

    ss = stats.fit([1, 2], distribution="uniform")
    tss = extract_input.from_neurom.transform_distr(ss)
    assert_equal(tss, {"uniform": {"min": 1, "max": 2}})

    ss = stats.fit([1, 2], distribution="expon")
    tss = extract_input.from_neurom.transform_distr(ss)
    assert_equal(tss, {"expon": {"loc": 1.0, "lambda": 2.0}})

    ss = stats.fit([1, 2], distribution="gamma")
    tss = extract_input.from_neurom.transform_distr(ss)
    assert_equal(tss, None)


def test_number_neurites(POPUL):
    res = extract_input.from_neurom.number_neurites(POPUL)
    assert_equal(
        res,
        {"num_trees": {"data": {"bins": [4, 5, 6, 7, 8, 9], "weights": [1, 0, 0, 0, 0, 1]}}},
    )


def test_number_neurites_cut_pop(POPUL):
    neurons = list(POPUL)
    if len(neurons[0].neurites) > len(neurons[1].neurites):
        smallest = 1
        biggest = 0
    else:
        smallest = 0
        biggest = 1

    for i in list(range(2, len(neurons[biggest].root_sections) - 1))[::-1]:
        neurons[biggest].delete_section(neurons[biggest].root_sections[i], recursive=True)

    POPUL = neurom.core.population.Population(neurons)
    assert_equal(len(neurons), 2)
    assert_equal(len(neurons[biggest].neurites), 3)
    assert_equal(len(neurons[smallest].neurites), 6)
    assert_equal(len(list(POPUL.neurites)), 9)
    res_cut = extract_input.from_neurom.number_neurites(POPUL)
    assert_equal(res_cut, {"num_trees": {"data": {"bins": [2, 3, 4], "weights": [1, 0, 1]}}})


def test_parameters():
    params = extract_input.parameters(
        neurite_types=["basal_dendrite", "apical_dendrite"],
        method="tmd",
        feature="radial_distances",
    )
    expected_params = {
        "basal_dendrite": {
            "randomness": 0.24,
            "targeting": 0.14,
            "radius": 0.3,
            "orientation": None,
            "growth_method": "tmd",
            "branching_method": "bio_oriented",
            "modify": None,
            "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
            "tree_type": 3,
            "metric": "radial_distances",
        },
        "apical_dendrite": {
            "randomness": 0.24,
            "targeting": 0.14,
            "radius": 0.3,
            "orientation": [[0.0, 1.0, 0.0]],
            "growth_method": "tmd_apical",
            "branching_method": "directional",
            "modify": None,
            "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
            "tree_type": 4,
            "metric": "radial_distances",
        },
        "axon": {},
        "origin": [0.0, 0.0, 0.0],
        "grow_types": ["basal_dendrite", "apical_dendrite"],
        "diameter_params": {"method": "default"},
    }
    assert_equal(params, expected_params)

    legacy_params = extract_input.parameters(
        neurite_types=["basal", "apical"],
        method="tmd",
        feature="radial_distances",
    )
    assert_equal(legacy_params, expected_params)

    default_params = extract_input.parameters(method="tmd", feature="radial_distances")
    expected_params["axon"] = {
        "randomness": 0.24,
        "targeting": 0.14,
        "radius": 0.3,
        "orientation": [[0.0, -1.0, 0.0]],
        "growth_method": "tmd",
        "branching_method": "bio_oriented",
        "modify": None,
        "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
        "metric": "radial_distances",
        "tree_type": 2,
    }
    expected_params["grow_types"].append("axon")
    assert_equal(default_params, expected_params)

    validator.validate_neuron_params(params)

    params_path = extract_input.parameters(
        neurite_types=["basal_dendrite", "apical_dendrite"], method="tmd"
    )

    assert_equal(
        params_path,
        {
            "basal_dendrite": {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": None,
                "growth_method": "tmd",
                "branching_method": "bio_oriented",
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "tree_type": 3,
                "metric": "path_distances",
            },
            "apical_dendrite": {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": [[0.0, 1.0, 0.0]],
                "growth_method": "tmd_apical",
                "branching_method": "directional",
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "tree_type": 4,
                "metric": "path_distances",
            },
            "axon": {},
            "origin": [0.0, 0.0, 0.0],
            "grow_types": ["basal_dendrite", "apical_dendrite"],
            "diameter_params": {"method": "default"},
        },
    )
    validator.validate_neuron_params(params_path)

    params_axon = extract_input.parameters(neurite_types=["axon"], method="tmd")

    assert_equal(
        params_axon,
        {
            "basal_dendrite": {},
            "apical_dendrite": {},
            "axon": {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": [[0.0, -1.0, 0.0]],
                "growth_method": "tmd",
                "branching_method": "bio_oriented",
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "tree_type": 2,
                "metric": "path_distances",
            },
            "origin": [0.0, 0.0, 0.0],
            "grow_types": ["axon"],
            "diameter_params": {"method": "default"},
        },
    )
    validator.validate_neuron_params(params_axon)

    params_axon = extract_input.parameters(neurite_types=["axon"], method="trunk")

    assert_equal(
        params_axon,
        {
            "basal_dendrite": {},
            "apical_dendrite": {},
            "axon": {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": [[0.0, -1.0, 0.0]],
                "growth_method": "trunk",
                "branching_method": "random",
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "tree_type": 2,
                "metric": "path_distances",
            },
            "origin": [0.0, 0.0, 0.0],
            "grow_types": ["axon"],
            "diameter_params": {"method": "default"},
        },
    )
    validator.validate_neuron_params(params)

    params_diameter = extract_input.parameters(
        neurite_types=["axon"], method="trunk", diameter_parameters="M1"
    )

    assert_equal(
        params_diameter,
        {
            "basal_dendrite": {},
            "apical_dendrite": {},
            "axon": {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": [[0.0, -1.0, 0.0]],
                "growth_method": "trunk",
                "branching_method": "random",
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "tree_type": 2,
                "metric": "path_distances",
            },
            "origin": [0.0, 0.0, 0.0],
            "grow_types": ["axon"],
            "diameter_params": {"method": "M1"},
        },
    )
    validator.validate_neuron_params(params_diameter)

    params_diameter_dict = extract_input.parameters(
        neurite_types=["axon"],
        method="trunk",
        diameter_parameters={"a": 1, "b": 2},
    )

    assert_equal(
        params_diameter_dict,
        {
            "basal_dendrite": {},
            "apical_dendrite": {},
            "axon": {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": [[0.0, -1.0, 0.0]],
                "growth_method": "trunk",
                "branching_method": "random",
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "tree_type": 2,
                "metric": "path_distances",
            },
            "origin": [0.0, 0.0, 0.0],
            "grow_types": ["axon"],
            "diameter_params": {"method": "external", "a": 1, "b": 2},
        },
    )
    validator.validate_neuron_params(params_diameter_dict)

    with pytest.raises(KeyError):
        extract_input.parameters(
            neurite_types=["axon"],
            method="UNKNOWN METHOD",
        )
    with pytest.raises(ValueError):
        extract_input.parameters(
            neurite_types=["axon"],
            method="trunk",
            diameter_parameters=object(),
        )


def test_from_TMD():
    files = sorted([os.path.join(POP_PATH, neuron_dir) for neuron_dir in os.listdir(POP_PATH)])
    pop = tmd.io.load_population(files)
    angles = extract_input.from_TMD.persistent_homology_angles(pop, neurite_type="basal_dendrite")
    expected = [
        [
            [39.8034782, 22.0587348, 0.2204977, -0.0031118, -0.7250693, 0.4353605],
            [47.0833969, 16.8049907, 0.4378265, 0.5057776, -1.4803076, 0.4328786],
            [142.8110504, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [31.5143489, 13.1089029, -0.0196220, 0.0082083, -0.8850177, 0.0621305],
            [31.6991310, 4.0347867, 0.5645827, -1.2206746, -1.958114, 0.3290293],
            [37.1221504, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [80.1259918, 45.2574195, 0.5985462, -0.7937724, -1.4556563, 0.2089598],
            [138.8324737, 137.2655334, 0.7664448, -0.5276853, -1.5705068, -0.1490920],
            [129.4999694, 58.3956718, -0.1864954, 0.2471917, -0.1825209, -0.8895693],
            [116.7485580, 29.0435256, 0.1938211, 0.8495636, -0.6798108, -0.3523395],
            [116.3598709, 4.6591954, -0.0381157, -0.5630145, -0.9937219, -0.0060555],
            [140.7718811, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [108.1560363, 24.4956016, 0.0795795, -0.2466572, -0.818391, 0.4079079],
            [84.4512786, 73.8114471, -0.0682486, -0.1316394, 0.7043639, 0.3319016],
            [41.1267166, 22.6543693, 0.3379619, -0.3496508, 0.2551283, 0.1817587],
            [58.6278800, 55.9588279, 0.1952034, -0.6733589, 0.0355827, 0.2554649],
            [67.1058807, 67.8727340, -0.0696913, -0.1445076, 4.92025, -0.8872188],
            [78.9355468, 16.0958881, -0.3179875, 0.5255199, -1.5740335, -0.2523424],
            [92.3746414, 50.6163444, 0.0482639, 0.1450342, 0.8074381, -0.2487332],
            [88.7393646, 14.1915130, -0.1820048, 0.6027378, 0.7695344, 0.4431591],
            [107.2000274, 13.7320985, 0.4757962, -1.473934, -1.7596798, 1.2384288],
            [157.3328552, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [124.3847961, 78.5777359, -0.2851194, 0.1851800, 0.7059138, -0.4451136],
            [67.1271133, 46.9302597, -0.0895297, 1.0650232, 6.1442575, -1.1252722],
            [50.9799156, 27.6712512, -5.344535, 0.164577, -0.0902247, 0.6043591],
            [147.5868072, 45.4994354, 0.0477559, -0.3307448, 0.5073497, -0.0292972],
            [125.9251632, 32.7475662, -0.3614119, 0.3678546, -0.4462253, -0.5532181],
            [119.1635360, 18.5245800, -0.0425105, 0.3792363, 0.3548733, -0.3755915],
            [115.0881805, 26.0076904, -0.1438069, -0.4749578, 5.3513756, -1.0207773],
            [44.8667869, 20.1308231, -0.5238513, 0.4554575, -0.3537357, 0.6082043],
            [124.8795852, 9.7746734, 0.1573835, -0.6259746, 0.5335315, 0.3545703],
            [159.7980194, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [73.5007705, 44.1673202, -0.2861464, 0.6478886, 0.7622488, -0.6312206],
            [187.1019897, 185.1600341, 0.5756854, -0.9131198, 4.854946, -0.0758894],
            [111.3322906, 23.4609394, -1.1881468, 1.4903252, -0.7528698, 0.3027869],
            [110.6484298, 24.8346958, 0.8869759, -0.5435520, 0.1794759, -1.2358185],
            [82.1550140, 12.0827064, 0.5742538, -0.2249454, 4.78082, 0.0218874],
            [187.2543487, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [128.7404022, 44.2018013, 0.1672787, 0.3934768, 0.4804887, -1.2018781],
            [202.8781890, 18.3915138, -0.4089744, 0.1219186, 5.5452833, 0.4808956],
            [219.8619384, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [224.0865325, 10.8910417, 0.0639219, -0.2720557, -0.536548, 0.8312207],
            [196.6136322, 2.6748218, -0.5220707, -0.7471831, 3.6071417, 0.5896363],
            [265.9921875, 0, np.nan, np.nan, np.nan, np.nan],
        ],
    ]
    for a, b in zip(angles["persistence_diagram"], expected):
        for ai, bi in zip(a, b):
            assert_array_almost_equal(ai, bi)

    angles = extract_input.from_TMD.persistent_homology_angles(
        pop, neurite_type="basal_dendrite", threshold=9
    )
    expected = [
        [
            [108.1560363, 24.4956016, 0.0795795, -0.2466572, -0.818391, 0.4079079],
            [84.4512786, 73.8114471, -0.0682486, -0.1316394, 0.7043639, 0.3319016],
            [41.1267166, 22.6543693, 0.3379619, -0.3496508, 0.2551283, 0.1817587],
            [58.6278800, 55.9588279, 0.1952034, -0.6733589, 0.0355827, 0.2554649],
            [67.1058807, 67.8727340, -0.0696913, -0.1445076, 4.92025, -0.8872188],
            [78.9355468, 16.0958881, -0.3179875, 0.5255199, -1.5740335, -0.2523424],
            [92.3746414, 50.6163444, 0.0482639, 0.1450342, 0.8074381, -0.2487332],
            [88.7393646, 14.1915130, -0.1820048, 0.6027378, 0.7695344, 0.4431591],
            [107.2000274, 13.7320985, 0.4757962, -1.473934, -1.7596798, 1.2384288],
            [157.3328552, 0, np.nan, np.nan, np.nan, np.nan],
        ],
        [
            [124.3847961, 78.5777359, -0.2851194, 0.1851800, 0.7059138, -0.4451136],
            [67.1271133, 46.9302597, -0.0895297, 1.0650232, 6.1442575, -1.1252722],
            [50.9799156, 27.6712512, -5.344535, 0.164577, -0.0902247, 0.6043591],
            [147.5868072, 45.4994354, 0.0477559, -0.3307448, 0.5073497, -0.0292972],
            [125.9251632, 32.7475662, -0.3614119, 0.3678546, -0.4462253, -0.5532181],
            [119.1635360, 18.5245800, -0.0425105, 0.3792363, 0.3548733, -0.3755915],
            [115.0881805, 26.0076904, -0.1438069, -0.4749578, 5.3513756, -1.0207773],
            [44.8667869, 20.1308231, -0.5238513, 0.4554575, -0.3537357, 0.6082043],
            [124.8795852, 9.7746734, 0.1573835, -0.6259746, 0.5335315, 0.3545703],
            [159.7980194, 0, np.nan, np.nan, np.nan, np.nan],
        ],
    ]
    for a, b in zip(angles["persistence_diagram"], expected):
        for ai, bi in zip(a, b):
            assert_array_almost_equal(ai, bi)



"""Definition of basic rotation functionality."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

import numpy as np

from neurots.morphmath.utils import norm


def spherical_from_vector(vect):
    """Return the spherical coordinates of a vector: phi, theta."""
    x, y, z = vect

    phi = np.arctan2(y, x)
    theta = np.arccos(z / norm(vect))

    return phi, theta


def vector_from_spherical(phi, theta):
    """Return a normalized vector from the spherical angles: phi, theta."""
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)

    return x, y, z


def rotation_around_axis(axis, angle):
    """Return a normalized vector rotated around the selected axis by an angle."""
    d = np.array(axis, dtype=float) / np.linalg.norm(axis)

    sn = np.sin(angle)
    cs = np.cos(angle)

    eye = np.eye(3, dtype=float)
    # ddt = np.outer(d, d)
    skew = np.array([[0, -d[2], d[1]], [d[2], 0, -d[0]], [-d[1], d[0], 0]], dtype=float)

    # mtx = ddt + cs * (eye - ddt) + sn * skew
    # mtx = cs * eye + sn * skew + (1. - cs) * ddt
    mtx = eye + sn * skew + (1.0 - cs) * np.linalg.matrix_power(skew, 2)
    return mtx


def angle3D(v1, v2):
    """Return the angle between v1, v2."""
    v1 = np.array(v1)
    v2 = np.array(v2)

    return math.acos(v1.dot(v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))


def rotate_vector(vec, axis, angle):
    """Rotate the input vector vec by a selected angle around a specific axis."""
    return np.dot(rotation_around_axis(axis, angle), vec)



"""Barcode utils."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np
from tmd.Topology.statistics import get_lengths
from tmd.Topology.transformations import tmd_scale

L = logging.getLogger(__name__)


def scale_barcode(ph, target_distance):
    """Scale a persistence homology from a given target distance.

    Given a target distance, scale the persistence homology in order to make sure that the neurite
    will reach the target.

    Args:
        ph (list[list]): The persistence homology.
        target_distance (float): The target distance.

    Returns:
        list[list]: The rescaled persistence homology.
    """
    ph_distance = np.nanmax(ph)

    if target_distance > ph_distance:
        ph = tmd_scale(ph, target_distance / ph_distance)

    return ph


def barcodes_greater_than_distance(ph_list, target_extent):
    """Returns all barcodes the max value of which is greater than target_extent.

    Args:
        ph_list (list[list[list]]): A list of persistence homologies.
        target_extent (float): The target barcode extent.

    Returns:
        list[list[list]]: The list of the selected barcodes.
    """
    max_extents = np.asarray([max(get_lengths(ph)) for ph in ph_list])
    mask = (max_extents > target_extent) | np.isclose(max_extents, target_extent)

    if not mask.any():
        L.warning("All barcodes are smaller than target. The longest is returned.")
        return [ph_list[np.argmax(max_extents)]]

    return [ph_list[index] for index in np.where(mask)[0]]



"""Test neurots.astrocyte.point_cloud code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import numpy as np
from numpy import testing as npt

from neurots.astrocyte.point_cloud import PointCloud


def point_array():
    return np.array(
        [
            [-10.0, -10.0, -10.0],  # 0
            [-8.0, -8.0, -8.0],  # 1
            [-6.0, -6.0, -6.0],  # 2
            [-4.0, -4.0, -4.0],  # 3
            [-2.0, -2.0, -2.0],  # 4
            [0.0, 0.0, 0.0],  # 5
            [2.0, 2.0, 2.0],  # 6
            [4.0, 4.0, 4.0],  # 7
            [6.0, 6.0, 6.0],  # 8
            [8.0, 8.0, 8.0],  # 9
            [10.0, 10.0, 10.0],  # 10
        ]
    )


def create_point_cloud():
    return PointCloud(point_array())


def test_constructor():
    point_cloud = create_point_cloud()
    npt.assert_allclose(point_cloud.points, point_array())

    n_points = len(point_cloud.points)

    npt.assert_array_equal(point_cloud.available_ids, np.arange(n_points, dtype=int))
    npt.assert_array_equal(point_cloud.removed_ids, [])
    npt.assert_array_equal(point_cloud.removed_points, np.empty((0, 3)))


def _assert_unordered_equal(arr1, arr2):
    npt.assert_array_equal(np.sort(arr1), np.sort(arr2))


def test_ball_query():

    point_cloud = create_point_cloud()

    ids = point_cloud.ball_query(np.array([1.0, 1.0, 1.0]), 0.001)
    npt.assert_array_equal(ids, [])

    ids = point_cloud.ball_query(np.array([1.0, 1.0, 1.0]), 2.0)

    _assert_unordered_equal(ids, [5, 6])


def test_upper_half_ball_query():

    point_cloud = create_point_cloud()

    ref_point = np.array([0.2, 0.2, 0.2])
    direction = -np.array([1.0, 1.0, 1.0])
    direction /= np.linalg.norm(direction)

    # ball_query will return [0, 1]
    ids = point_cloud.upper_half_ball_query(ref_point, 10.0, direction)

    _assert_unordered_equal(ids, [3, 4, 5])


def test_partial_ball_query():

    point_cloud = create_point_cloud()

    ref_point = -np.ones(3)
    direction = np.array([1.0, 1.0, 1.0])
    direction /= np.linalg.norm(direction)

    radius = np.linalg.norm(ref_point - np.array([5.0, 5.0, 5.0]))

    d1 = np.linalg.norm(ref_point - np.array([2.5, 2.5, 2.5]))
    d2 = np.linalg.norm(ref_point - np.array([-4.1, -4.1, -4.0]))

    cap_angle_front = np.arccos(d1 / radius)
    cap_angle_back = -np.arccos(d2 / radius)

    # ball_query will return [0, 1]
    ids = point_cloud.partial_ball_query(
        ref_point, radius, direction, cap_angle_front, cap_angle_back
    )

    _assert_unordered_equal(ids, [3, 4, 5, 6])


def test_remove_ids():

    to_remove = np.array([4, 3, 9, 1])

    point_cloud = create_point_cloud()
    point_cloud.remove_ids(to_remove)

    npt.assert_array_equal(point_cloud.removed_ids, [1, 3, 4, 9])


def test_nearest_neighbor_direction():

    point_cloud = create_point_cloud()

    point = np.ones(3) + 0.2

    expected_direction = np.ones(3) * 2.0 - point
    expected_direction /= np.linalg.norm(expected_direction)

    npt.assert_allclose(point_cloud.nearest_neighbor_direction(point, 2.0), expected_direction)

    assert point_cloud.nearest_neighbor_direction(point, 0.001) is None


def test_remove_points_around():
    point_cloud = create_point_cloud()
    point = np.array([0.1, 0.1, 0.1])

    all_ids = point_cloud.available_ids

    removed_ids_1 = point_cloud.remove_points_around(point, 0.2)

    assert np.intersect1d(removed_ids_1, point_cloud.available_ids).size == 0

    remaining_ids = np.setdiff1d(all_ids, removed_ids_1)
    npt.assert_array_equal(remaining_ids, point_cloud.available_ids)

    removed_ids_2 = point_cloud.remove_points_around(point, 0.3)
    assert np.intersect1d(removed_ids_2, point_cloud.available_ids).size == 0

    remaining_ids = np.setdiff1d(remaining_ids, removed_ids_2)
    npt.assert_array_equal(remaining_ids, point_cloud.available_ids)


def test_remove_hemisphere():
    point_cloud = create_point_cloud()

    point = np.zeros(3)
    radius = 100.0
    direction = -np.ones(3)
    direction /= np.linalg.norm(direction)

    point_cloud.remove_hemisphere(point, direction, radius)

    npt.assert_array_equal(point_cloud.removed_ids, [0, 1, 2, 3, 4, 5])



"""Definition of bifurcation functionality."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

import neurots.morphmath.rotation as rt
from neurots.morphmath.utils import get_random_point


def random(random_generator=np.random):
    """Get 3-d coordinates of a new random point.

    The distance between the produced point and (0,0,0) is given by the value D.
    """
    dir1 = get_random_point(random_generator=random_generator)
    dir2 = get_random_point(random_generator=random_generator)

    return (np.array(dir1), np.array(dir2))


def symmetric(direction, angles):
    """Get 3-d coordinates for two new directions at a selected angle."""
    # phi0 = angles[0] #not used
    # theta0 = angles[1] #not used
    phi1 = angles[2] / 2.0
    theta1 = angles[3] / 2.0

    dir1 = rt.rotate_vector(direction, [0, 0, 1], phi1)
    dir1 = rt.rotate_vector(dir1, [1, 0, 0], theta1)
    dir2 = rt.rotate_vector(direction, [0, 0, 1], -phi1)
    dir2 = rt.rotate_vector(dir2, [1, 0, 0], -theta1)

    return (np.array(dir1), np.array(dir2))


def bio_oriented(direction, angles):
    """Input: init_phi, init_theta, dphi, dtheta."""
    phi0 = angles[0]
    theta0 = angles[1]
    phi1 = angles[2]
    theta1 = angles[3]

    dir1 = rt.rotate_vector(direction, [0, 0, 1], phi0)
    dir1 = rt.rotate_vector(dir1, [1, 0, 0], theta0)
    dir2 = rt.rotate_vector(dir1, [0, 0, 1], phi1)
    dir2 = rt.rotate_vector(dir2, [1, 0, 0], theta1)

    return (np.array(dir1), np.array(dir2))


def directional(direction, angles):
    """Input: init_phi, init_theta, dphi, dtheta."""
    # phi0 = angles[0] #not used
    # theta0 = angles[1] #not used
    phi1 = angles[2]
    theta1 = angles[3]

    dir2 = rt.rotate_vector(direction, [0, 0, 1], phi1)
    dir2 = rt.rotate_vector(dir2, [1, 0, 0], theta1)

    return (np.array(direction), np.array(dir2))



"""Test the neurots.generate.grower.NeuronGrower class.

This test ensures that the radial and path distances are computed correctly through NeuroTS,
so that the code is treating the input barcode, according to the given parameters.
For this reason, we need to check that the same input distribution
will generate cells with different properties, according to their input parameters.
Finally, we need to check the TMD of the produced cells.
"""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=protected-access
import json
import os
from os.path import join
from tempfile import TemporaryDirectory

import numpy as np
import pytest
import tmd
from morph_tool import diff
from neurom.core import Morphology
from numpy.testing import assert_almost_equal
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_array_equal
from scipy.spatial.distance import cdist

from neurots.generate.diametrizer import diametrize_constant_per_neurite
from neurots.generate.grower import NeuronGrower

_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def build_random_generator(seed=None):
    mt = np.random.MT19937()
    mt._legacy_seeding(seed)  # Use legacy seeding to get the same result as with np.random.seed()
    return np.random.RandomState(mt)


def assert_close_persistent_diagram(actual, expected):
    # compute distances between points
    distances = np.min(cdist(np.array(tmd.analysis.sort_ph(expected)), actual), axis=0)
    # We compare distances between expected and generated peristence as it is more stable to check.
    # This comparison does not depend on ordering of the points
    # and ensures that the points of the original persistence are consistently generated.
    assert_almost_equal(len(expected), len(actual))
    assert_array_almost_equal(distances, np.zeros(len(distances)), decimal=0.1)
    assert_almost_equal(np.max(expected[-1]), np.max(actual[-1]), decimal=0.1)


def _load_inputs(distributions, parameters):
    with open(distributions, encoding="utf-8") as f:
        distributions = json.load(f)

    with open(parameters, encoding="utf-8") as f:
        params = json.load(f)

    return distributions, params


def _test_full(
    feature,
    distributions,
    parameters,
    ref_cell,
    ref_persistence_diagram,
    save=False,
    rng_or_seed=None,
    skip_validation=False,
):
    distributions, params = _load_inputs(join(_path, distributions), join(_path, parameters))
    if rng_or_seed is None:
        np.random.seed(0)
        n = NeuronGrower(
            input_distributions=distributions,
            input_parameters=params,
            skip_validation=skip_validation,
        ).grow()
    else:
        n = NeuronGrower(
            input_distributions=distributions,
            input_parameters=params,
            rng_or_seed=rng_or_seed,
            skip_validation=skip_validation,
        ).grow()

    with TemporaryDirectory("test_grower") as folder:
        out_neuron = os.path.join(folder, "test_output_neuron_.h5")
        n.write(out_neuron)

        # For checking purposes, we can output the cells as swc
        if save:
            n.write(ref_cell.replace(".h5", "NEW.h5"))

        if ref_persistence_diagram is not None:
            # Load with TMD and extract radial persistence
            n0 = tmd.io.load_neuron(out_neuron)

            actual_persistence_diagram = tmd.methods.get_persistence_diagram(
                n0.apical_dendrite[0], feature=feature
            )
            if save:
                print(actual_persistence_diagram)

            with open(join(_path, ref_persistence_diagram), encoding="utf-8") as f:
                expected_persistence_diagram = json.load(f)

            assert_close_persistent_diagram(
                actual_persistence_diagram, expected_persistence_diagram
            )

        assert not diff(out_neuron, os.path.join(_path, ref_cell))


def test_wrong_filtration():
    """Test filtration metric inconsistency in distrib and params: path != radial"""
    distributions, parameters = _load_inputs(
        os.path.join(_path, "bio_path_distribution.json"),
        os.path.join(_path, "bio_radial_params.json"),
    )
    with pytest.raises(ValueError):
        NeuronGrower(parameters, distributions)


def test_seeding():
    """Test seeding of internal random number generator"""
    distributions, parameters = _load_inputs(
        os.path.join(_path, "bio_path_distribution.json"),
        os.path.join(_path, "bio_path_params.json"),
    )
    ng = NeuronGrower(parameters, distributions, rng_or_seed=0)
    assert ng._rng.bit_generator.state == {
        "bit_generator": "PCG64",
        "state": {
            "state": 80186449399738619878794082838194943960,
            "inc": 87136372517582989555478159403783844777,
        },
        "has_uint32": 0,
        "uinteger": 0,
    }

    ng = NeuronGrower(parameters, distributions, rng_or_seed=None)
    assert ng._rng.bit_generator.state["bit_generator"] == "PCG64"


def test_grow_trunk_1_basal():
    """Test NeuronGrower._grow_trunk() with only 1 basal (should raise an Exception)"""
    distributions, parameters = _load_inputs(
        os.path.join(_path, "bio_path_distribution.json"),
        os.path.join(_path, "bio_path_params.json"),
    )
    distributions["basal_dendrite"]["num_trees"]["data"]["bins"] = [1]
    ng = NeuronGrower(parameters, distributions)
    with pytest.raises(Exception, match=r"There should be at least 2 basal dendrites \(got 1\)"):
        ng.grow()

    parameters["basal_dendrite"]["orientation"] = {
        "mode": "use_predefined",
        "values": {"orientations": [[0.0, 1.0, 0.0]]},
    }
    ng = NeuronGrower(parameters, distributions)
    with pytest.raises(Exception, match=r"There should be at least 2 basal dendrites \(got 1\)"):
        ng.grow()


def test_external_diametrizer():
    """Test external diametrizer"""
    distributions, parameters = _load_inputs(
        os.path.join(_path, "bio_path_distribution.json"),
        os.path.join(_path, "bio_path_params.json"),
    )
    distributions["diameter"]["method"] = "M1"
    parameters["diameter_params"]["method"] = "M1"
    ng = NeuronGrower(parameters, distributions, rng_or_seed=0)
    ng.grow()

    # Inconsistent methods
    distributions["diameter"]["method"] = "external"
    parameters["diameter_params"]["method"] = "M1"
    with pytest.raises(
        ValueError,
        match="Diameters methods of parameters and distributions is inconsistent: M1 != external",
    ):
        NeuronGrower(parameters, distributions)

    # No external diametrizer provided
    distributions["diameter"]["method"] = "external"
    parameters["diameter_params"]["method"] = "external"
    with pytest.raises(
        ValueError, match="External diametrizer is missing the diametrizer function."
    ):
        NeuronGrower(parameters, distributions)

    bad_ng = NeuronGrower(parameters, distributions, external_diametrizer=object())
    with pytest.raises(Exception, match="Please provide an external diametrizer!"):
        bad_ng._init_diametrizer()

    # Test with an external diametrizer and neurite_types in diameter_params
    distributions["diameter"]["method"] = "external"
    parameters["diameter_params"]["method"] = "external"
    parameters["diameter_params"]["neurite_types"] = parameters["grow_types"]
    ng_external = NeuronGrower(
        parameters,
        distributions,
        external_diametrizer=diametrize_constant_per_neurite,
        rng_or_seed=0,
    )
    ng_external.grow()

    assert (Morphology(ng.neuron).points == Morphology(ng_external.neuron).points).all()


def test_convert_orientation2points():
    """Test NeuronGrower._convert_orientation2points()"""
    np.random.seed(0)
    distributions, parameters = _load_inputs(
        os.path.join(_path, "bio_path_distribution.json"),
        os.path.join(_path, "bio_path_params.json"),
    )
    distributions["diameter"]["method"] = "M1"
    parameters["diameter_params"]["method"] = "M1"
    ng = NeuronGrower(parameters, distributions)

    pts = ng._convert_orientation2points([[0, 1, 0]], 1, distributions["apical_dendrite"], {})
    assert_array_almost_equal(pts, [[0, 15.27995, 0]])

    # Test with no existing trunk
    ng = NeuronGrower(parameters, distributions)
    pts = ng._convert_orientation2points(None, 2, distributions["apical_dendrite"], {})
    assert_array_almost_equal(
        pts, [[-10.399604, -0.173343, 0.937449], [10.31932, 0.172005, -1.594578]]
    )

    with pytest.raises(ValueError):
        ng._convert_orientation2points("from_space", 1, distributions["apical_dendrite"], {})

    # Test with existing trunks
    ng.grow()
    pts = ng._convert_orientation2points(None, 2, distributions["apical_dendrite"], {})

    assert_array_almost_equal(pts, [[2.770599, 4.868847, 8.813554], [-6.314678, 6.2103, 5.533321]])

    with pytest.raises(ValueError):
        ng._convert_orientation2points(object(), 1, distributions["apical_dendrite"], {})

    with pytest.raises(ValueError):
        ng._convert_orientation2points([[0, 1, 0]], 99, distributions["apical_dendrite"], {})

    distributions, parameters = _load_inputs(
        os.path.join(_path, "axon_trunk_distribution.json"),
        os.path.join(_path, "axon_trunk_parameters_absolute.json"),
    )
    with pytest.raises(ValueError):
        ng._convert_orientation2points(
            [[0, 1, 0], [0, 1, 0]],
            1,
            distributions["axon"],
            {"trunk_absolute_orientation": True},
        )

    del distributions["axon"]["trunk"]["absolute_elevation_deviation"]
    with pytest.raises(KeyError):
        ng._convert_orientation2points(
            [[0, 1, 0]], 1, distributions["axon"], {"trunk_absolute_orientation": True}
        )


def test_breaker_of_tmd_algo():
    """Test example that could break tmd_algo. Test should fail if problem occurs.
    Otherwise, this grower should run smoothly.
    """
    distributions, params = _load_inputs(
        join(_path, "bio_distr_breaker.json"), join(_path, "bio_params_breaker.json")
    )
    np.random.seed(3367155)
    N = NeuronGrower(input_distributions=distributions, input_parameters=params)
    n = N.grow()

    assert_array_equal(N.apical_sections, [33])
    assert_array_almost_equal(
        n.sections[169].points[-1],
        np.array([-220.93813, -21.49141, -55.93323]),
        decimal=5,
    )
    assert_array_almost_equal(
        n.sections[122].points[-1], np.array([-17.31787, 151.4876, -6.67741]), decimal=5
    )

    # Test with a specific random generator
    rng = build_random_generator(3367155)

    N = NeuronGrower(input_distributions=distributions, input_parameters=params, rng_or_seed=rng)
    n = N.grow()

    assert_array_equal(N.apical_sections, [33])
    assert_array_almost_equal(
        n.sections[169].points[-1],
        np.array([-220.93813, -21.49141, -55.93323]),
        decimal=5,
    )
    assert_array_almost_equal(
        n.sections[122].points[-1], np.array([-17.31787, 151.4876, -6.67741]), decimal=5
    )


def test_axon_grower():
    """Test axon grower, which should only grow trunks with 1 section to allow later axon grafting.

    The num_seg value in the parameters is set to 999 but only 1 segment should be synthesized.
    """

    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters.json",
        "test_axon_grower.h5",
        None,
    )
    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters.json",
        "test_axon_grower.h5",
        None,
        rng_or_seed=build_random_generator(0),
    )
    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters_orientation_manager.json",
        "test_axon_grower.h5",
        None,
        skip_validation=True,
    )
    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters_orientation_manager.json",
        "test_axon_grower.h5",
        None,
        rng_or_seed=build_random_generator(0),
        skip_validation=True,
    )

    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters_absolute.json",
        "test_axon_grower_absolute.h5",
        None,
    )
    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters_absolute.json",
        "test_axon_grower_absolute.h5",
        None,
        rng_or_seed=build_random_generator(0),
    )
    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters_absolute_orientation_manager.json",
        "test_axon_grower_absolute.h5",
        None,
        skip_validation=True,
    )
    _test_full(
        "radial_distances",
        "axon_trunk_distribution.json",
        "axon_trunk_parameters_absolute_orientation_manager.json",
        "test_axon_grower_absolute.h5",
        None,
        rng_or_seed=build_random_generator(0),
        skip_validation=True,
    )


def test_basic_grower():
    _test_full(
        "radial_distances",
        "bio_trunk_distribution.json",
        "trunk_parameters.json",
        "test_trunk_grower.h5",
        None,
    )
    _test_full(
        "radial_distances",
        "bio_trunk_distribution.json",
        "trunk_parameters.json",
        "test_trunk_grower.h5",
        None,
        rng_or_seed=build_random_generator(0),
    )


def test_basic_grower_with_generator():
    distributions, params = _load_inputs(
        join(_path, "bio_trunk_distribution.json"),
        join(_path, "trunk_parameters.json"),
    )
    expected_pts = [
        [-0.7312348484992981, 7.604228973388672, 11.173797607421875],
        [-13.377432823181152, -1.2863954305648804, 2.9336819648742676],
        [11.861421585083008, -0.049414388835430145, 6.1279988288879395],
        [-2.3804218769073486, 12.54181957244873, 1.118072748184204],
    ]

    rng = np.random.default_rng(0)
    rng_or_seeds = [0, rng]

    for rng_or_seed in rng_or_seeds:
        n = NeuronGrower(
            input_distributions=distributions,
            input_parameters=params,
            rng_or_seed=rng_or_seed,
        ).grow()
        assert len(n.root_sections) == 4
        assert_array_almost_equal(
            [i.points[-1].tolist() for i in n.root_sections],
            expected_pts,
        )

    with pytest.raises(TypeError):
        NeuronGrower(params, distributions, rng_or_seed="NOT A SEED")


class TestPathGrower:
    """test tmd_path and tmd_apical_path"""

    def test_default(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_path_params.json",
            "path_grower.h5",
            "bio_path_persistence_diagram.json",
        )

    def test_with_rng(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_path_params.json",
            "path_grower.h5",
            "bio_path_persistence_diagram.json",
            rng_or_seed=build_random_generator(0),
        )

    def test_skip_validation(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_path_params_orientation_manager.json",
            "path_grower.h5",
            "bio_path_persistence_diagram.json",
            skip_validation=True,
        )

    def test_skip_rng_and_validation(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_path_params_orientation_manager.json",
            "path_grower.h5",
            "bio_path_persistence_diagram.json",
            skip_validation=True,
            rng_or_seed=build_random_generator(0),
        )


class TestGradientPathGrower:
    """test tmd_path"""

    def test_default(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_gradient_path_params.json",
            "gradient_path_grower.h5",
            "gradient_path_persistence_diagram.json",
        )

    def test_with_rng(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_gradient_path_params.json",
            "gradient_path_grower.h5",
            "gradient_path_persistence_diagram.json",
            rng_or_seed=build_random_generator(0),
        )

    def test_skip_validation(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_gradient_path_params_orientation_manager.json",
            "gradient_path_grower.h5",
            "gradient_path_persistence_diagram.json",
            skip_validation=True,
        )

    def test_skip_rng_and_validation(self):
        _test_full(
            "path_distances",
            "bio_distribution.json",
            "bio_gradient_path_params_orientation_manager.json",
            "gradient_path_grower.h5",
            "gradient_path_persistence_diagram.json",
            skip_validation=True,
            rng_or_seed=build_random_generator(0),
        )


class TestBioRatL5Tpc1:
    def test_default(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params1.json",
            "expected_bio_rat_L5_TPC_B_with_params1.h5",
            "expected_bio_rat_L5_TPC_B_with_params1_persistence_diagram.json",
        )

    def test_with_rng(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params1.json",
            "expected_bio_rat_L5_TPC_B_with_params1.h5",
            "expected_bio_rat_L5_TPC_B_with_params1_persistence_diagram.json",
            rng_or_seed=build_random_generator(0),
        )

    def test_skip_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params1_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params1.h5",
            "expected_bio_rat_L5_TPC_B_with_params1_persistence_diagram.json",
            skip_validation=True,
        )

    def test_skip_rng_and_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params1_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params1.h5",
            "expected_bio_rat_L5_TPC_B_with_params1_persistence_diagram.json",
            skip_validation=True,
            rng_or_seed=build_random_generator(0),
        )


class TestBioRatL5Tpc2:
    def test_default(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params2.json",
            "expected_bio_rat_L5_TPC_B_with_params2.h5",
            "expected_bio_rat_L5_TPC_B_with_params2_persistence_diagram.json",
        )

    def test_with_rng(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params2.json",
            "expected_bio_rat_L5_TPC_B_with_params2.h5",
            "expected_bio_rat_L5_TPC_B_with_params2_persistence_diagram.json",
            rng_or_seed=build_random_generator(0),
        )

    def test_skip_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params2_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params2.h5",
            "expected_bio_rat_L5_TPC_B_with_params2_persistence_diagram.json",
            skip_validation=True,
        )

    def test_skip_rng_and_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params2_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params2.h5",
            "expected_bio_rat_L5_TPC_B_with_params2_persistence_diagram.json",
            skip_validation=True,
            rng_or_seed=build_random_generator(0),
        )


class TestBioRatL5Tpc3:
    def test_default(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params3.json",
            "expected_bio_rat_L5_TPC_B_with_params3.h5",
            "expected_bio_rat_L5_TPC_B_with_params3_persistence_diagram.json",
        )

    def test_with_rng(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params3.json",
            "expected_bio_rat_L5_TPC_B_with_params3.h5",
            "expected_bio_rat_L5_TPC_B_with_params3_persistence_diagram.json",
            rng_or_seed=build_random_generator(0),
        )

    def test_skip_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params3_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params3.h5",
            "expected_bio_rat_L5_TPC_B_with_params3_persistence_diagram.json",
            skip_validation=True,
        )

    def test_skip_rng_and_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params3_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params3.h5",
            "expected_bio_rat_L5_TPC_B_with_params3_persistence_diagram.json",
            skip_validation=True,
            rng_or_seed=build_random_generator(0),
        )


class TestBioRatL5Tpc4:
    def test_default(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params4.json",
            "expected_bio_rat_L5_TPC_B_with_params4.h5",
            "expected_bio_rat_L5_TPC_B_with_params4_persistence_diagram.json",
        )

    def test_with_rng(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params4.json",
            "expected_bio_rat_L5_TPC_B_with_params4.h5",
            "expected_bio_rat_L5_TPC_B_with_params4_persistence_diagram.json",
            rng_or_seed=build_random_generator(0),
        )

    def test_skip_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params4_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params4.h5",
            "expected_bio_rat_L5_TPC_B_with_params4_persistence_diagram.json",
            skip_validation=True,
        )

    def test_skip_rng_and_validation(self):
        _test_full(
            "path_distances",
            "bio_rat_L5_TPC_B.json",
            "params4_orientation_manager.json",
            "expected_bio_rat_L5_TPC_B_with_params4.h5",
            "expected_bio_rat_L5_TPC_B_with_params4_persistence_diagram.json",
            skip_validation=True,
            rng_or_seed=build_random_generator(0),
        )



"""Space colonization context data structures."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np
from scipy.special import logit

from neurots.astrocyte.point_cloud import PointCloud
from neurots.morphmath.point_array import DynamicPointArray
from neurots.utils import NeuroTSError

L = logging.getLogger(__name__)


class EndfeetTargets:
    """Store the Endfeet target points.

    Endfeet targets are the 3D points that influence the growth of the targeted tree growth.
    A tree will grow towards these points as long as they remain active.
    If a point is deactivated then it will no longer affect the grower it is assigned to.

    Args:
        coordinates (numpy.ndarray):
            Array of 3D points.

    Attributes:
        points (numpy.ndarray[numpy.single]):
            The coordinates of the 3D points that represent the endfeet
            targets.
        active (numpy.ndarray[bool]):
            A boolean mask, the same size as the number of points. A point
            is active when its corresponding boolean value in the active array
            is True.
    """

    def __init__(self, coordinates):
        self.points = np.asarray(coordinates, dtype=np.float32)
        self.active = np.ones(len(coordinates), dtype=bool)

    def __len__(self):
        return len(self.active)

    @property
    def active_points(self):
        """Returns the points that are still active."""
        return self.points[self.active]


class SpaceColonizationContext:
    """Context class for space colonization. It includes globally available information.

    Attributes:
        morphology_points (DynamicPointArray):
            The morphology points from the entire morphology, globally available by algorithms
            that need access to the morphology as a whole. For example repulsion needs to have
            access to all the neighboring points in the morphology to calculate the repulsion
            from them.
        endfeet_targets (EndfeetTargets):
            The targets for the targeting space colonization algorithm.
        field (Callable[float] -> float):
            The attraction field function that specifies how strong is the field as we approach
            the target. It takes a fraction which represents the ratio of distance of a point to
            the target over the distance from the soma to the target. That means that it will be
            zero when overlapping with the target and one if the point is overlapping with the
            soma position. It is used by the targeting space colonization algorithm to determine
            how much influenced the splitting direction should by the presence of the target.
        point_cloud (PointCloud):
            Seed point cloud for the space colonization queries.
        collision_handle (Callable[numpy.ndarray, float] -> bool):
            A callable function that takes a point and the segment length as input an returns
            if there is a collision or not. The segment length is used by probabilistic checks
            that scale the probability of colliding (soft collision with exponential decay from
            boundary distance), because it would affect the probability if a check is made
            every 1.0 um or every 0.1 um (10x more times).

    """

    def __init__(self, params):

        self.morphology_points = DynamicPointArray()

        if "endfeet_targets" in params:
            self.endfeet_targets = EndfeetTargets(params["endfeet_targets"])
        else:
            L.info("No endfeet targets available")
            self.endfeet_targets = None

        field = params["field"]
        if field["type"] == "logit":
            self.field = lambda x: field["slope"] * logit(x) + field["intercept"]
        else:
            raise NeuroTSError(f"{field['type']} function type is not available")

        sc_params = params["space_colonization"]
        self._params = sc_params

        if "point_cloud" in sc_params:
            self.point_cloud = PointCloud(sc_params["point_cloud"])
        else:
            raise NeuroTSError("point_cloud entry is not available in params")

        if "collision_handle" not in params or params["collision_handle"] is None:
            self.collision_handle = lambda *args: False
            L.info("No collision handle provided. There will be no collision checks.")
        else:
            self.collision_handle = params["collision_handle"]

    def kill_distance(self, segment_length):
        """Space colonization algorithm kill distance.

        Space colonization requires the removal of seed points (point cloud) around the newly
        created points. Kill distance is the threshold radius of the ball query to find the points
        to remove.

        Note:
            Instead of setting an absolute number it makes more sense for the
            kill distance to be factor multiplied by the segment length.

        For more details see:
        http://algorithmicbotany.org/papers/colonization.egwnp2007.large.pdf
        """
        kill_distance_factor = self._params["kill_distance_factor"]
        return kill_distance_factor * segment_length

    def influence_distance(self, segment_length):
        """Space colonization algorithm influence distance.

        Space colonization requires finding the seeds (point cloud) around points that will
        influence the growth direction. Influence distance is the threshold radius of the
        ball query to find the points to contribute.

        Note:
            Instead of setting an absolute number it makes more sense for the
            influence distance to be factor multiplied by the segment length.

        For more details see:
        http://algorithmicbotany.org/papers/colonization.egwnp2007.large.pdf
        """
        influence_distance_factor = self._params["influence_distance_factor"]
        return influence_distance_factor * segment_length



"""Basic class for TreeGrower Algorithms."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import logging

import numpy as np

from neurots.generate.algorithms.abstractgrower import AbstractAlgo
from neurots.generate.algorithms.barcode import Barcode
from neurots.generate.algorithms.common import TMDStop
from neurots.generate.algorithms.common import bif_methods
from neurots.generate.algorithms.common import section_data
from neurots.morphmath import sample
from neurots.morphmath.utils import norm

L = logging.getLogger(__name__)


class TMDAlgo(AbstractAlgo):
    """TreeGrower of TMD basic growth.

    Args:
        input_data (dict): All the data required for the growth.
        params (dict): The parameters required for growth. It should include the
            ``branching_method`` selected from: ``|bio_oriented, symmetric, directional]``.
        start_point (list[float]): The first point of the trunk.
        skip_validation (bool): If set to ``False``, the parameters and distributions and
            the "min_bar_length" parameter are validated.
        context (Any): An object containing contextual information.
        random_generator (numpy.random.Generator): The random number generator to use.
    """

    def __init__(
        self,
        input_data,
        params,
        start_point,
        skip_validation=False,
        context=None,
        random_generator=np.random,
        **_,
    ):
        """TMD basic grower."""
        super().__init__(input_data, params, start_point, context)
        self.bif_method = bif_methods[params["branching_method"]]
        self.params = copy.deepcopy(params)
        self.ph_angles = self.select_persistence(input_data, random_generator)
        self.barcode = Barcode(list(self.ph_angles))
        self.start_point = start_point
        self.apical_section = None
        self.apical_point_distance_from_soma = 0.0
        self.persistence_length = self.barcode.get_persistence_length()
        # Validate parameters and distributions
        if not skip_validation:
            # Consistency check between parameters - persistence diagram
            barSZ = input_data["min_bar_length"]
            stepSZ = params["step_size"]["norm"]["mean"]
            if stepSZ >= barSZ:
                L.warning(
                    "Selected step size %f is too big for bars of size %f",
                    stepSZ,
                    barSZ,
                )

    def select_persistence(self, input_data, random_generator=np.random):
        """Select the persistence.

        Sample one persistence diagram from a list of diagrams and modifies according to input
        parameters.
        """
        list_of_persistences = input_data["persistence_diagram"]
        persistence = sample.ph(list_of_persistences, random_generator)

        if self.params.get("modify"):
            persistence = self.params["modify"]["funct"](
                persistence, self.context, **self.params["modify"]["kwargs"]
            )
        return persistence

    @staticmethod
    def metric_ref(section):
        """Return the metric reference.

        The metric reference is the path distance reference, or zero if no section is provided as
        input.
        """
        # Function to return reference for path distance
        if section:
            return section.pathlength
        return 0.0

    @staticmethod
    def metric(section):
        """Return the metric at the current position, here path distance, recorded in section."""
        # Function to return path distance
        return section.pathlength

    def get_stop_criteria(self, current_section):
        """Return stop criteria that are the commonly shared by all TMDPath algorithms.

        Returns:
            tuple[dict, dict]: Two dictionaries, each containing one entry whose key is ``TMD`` and
            value is a :class:`neurots.generate.algorithms.common.TMDStop` object.
        """
        # Ensure that reference is correctly assigned
        current_section.stop_criteria["TMD"].ref = self.metric_ref(current_section)
        # Copy the values for the parent stop TMD to use
        parent_tmd = copy.deepcopy(current_section.stop_criteria["TMD"])
        # Save the values of bifurcation for parent
        parent_bif_id = parent_tmd.bif_id
        parent_bif = parent_tmd.bif
        # Define the current criterion, inherited from parent
        current_tmd = copy.deepcopy(current_section.stop_criteria["TMD"])

        # The termination remains the same, so it is always True that
        # current_tmd.term <= parent_tmd.term
        # We find the smallest bifurcation that fulfils requirements
        # parent_tmd.ref <= current_tmd.bif <= parent_tmd.term
        # Bifurcation is larger than current reference distance
        bif_id, bif = self.barcode.min_bif(bif_above=parent_tmd.ref, bif_below=parent_tmd.term)
        # Update the bifurcation in the stop_criterion
        current_tmd.update_bif(bif_id, bif)
        # Ensure that criterion fulfils all requirements
        # the term that corresponds to current_tmd.bif term_target
        # term_target <= parent_tmd.term
        # If not re-assign a new one, find the min bifurcation for which:
        # term_target <= parent_tmd.term
        target_stop1 = self.barcode.curate_stop_criterion(parent_tmd, current_tmd)
        stop1 = {"TMD": target_stop1}
        # Find the termination that fulfils the requirement
        # termination <= current termination

        # Use the current bifurcation to determine the respective termination
        # Bifurcation should be larger than current reference distance
        term_id, term = self.barcode.get_term_between(parent_bif_id, parent_bif, current_tmd.term)
        current_tmd.update_term(term_id, term)

        # Get a stop criterion that fulfils requirements
        target_stop2 = self.barcode.curate_stop_criterion(parent_tmd, current_tmd)
        stop2 = {"TMD": target_stop2}

        return (stop1, stop2)

    def initialize(self):
        """Generates the data to be used for the initialization of the first section to be grown.

        Saves the extracted input data into the corresponding structures.

        Returns:
            tuple[TMDStop, int]: A :class:`neurots.generate.algorithms.common.TMDStop` object and
            the number of sections.
        """
        b0_id, b0 = self.barcode.min_bif()
        t0_id, t0 = self.barcode.max_term()
        stop = {
            "TMD": TMDStop(ref=self.metric_ref(None), bif_id=b0_id, bif=b0, term_id=t0_id, term=t0)
        }

        num_sec = len(self.ph_angles)

        return stop, num_sec

    def bifurcate(self, current_section):
        """When the section bifurcates two new sections need to be created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.

        Returns:
            tuple[dict, dict]: Two dictionaries containing the data of the two children sections.
        """
        self.barcode.remove_bif(current_section.stop_criteria["TMD"].bif_id)
        ang = self.barcode.angles[current_section.stop_criteria["TMD"].bif_id]

        dir1, dir2 = self.bif_method(current_section.history(), angles=ang)
        first_point = np.array(current_section.last_point)

        stop1, stop2 = self.get_stop_criteria(current_section)

        return (
            section_data(dir1, first_point, stop1, current_section.process),
            section_data(dir2, first_point, stop2, current_section.process),
        )

    def terminate(self, current_section):
        """Terminate the current section.

        When the growth of a section is terminated the "term" must be removed from the TMD grower.
        """
        self.barcode.remove_term(current_section.stop_criteria["TMD"].term_id)

    def extend(self, current_section):
        """Definition of stop criterion for the growth of the current section."""
        criteria_tmd = copy.deepcopy(current_section.stop_criteria["TMD"])
        maximum_target = current_section.stop_criteria["TMD"].term
        reference = current_section.stop_criteria["TMD"].ref

        # We check that the current bifurcation has not been used
        if criteria_tmd.bif_id not in self.barcode.bifs and not np.isinf(criteria_tmd.bif):
            criteria_tmd.update_bif(
                *self.barcode.min_bif(bif_above=reference, bif_below=maximum_target)
            )
            criteria_tmd = self.barcode.curate_stop_criterion(criteria_tmd, criteria_tmd)

        # We check that the current termination has not been used
        if criteria_tmd.term_id not in self.barcode.terms:
            # Termination must be larger that bifurcation
            # unless if bifurcation is infinite
            reference = criteria_tmd.bif if not np.isinf(criteria_tmd.bif) else criteria_tmd.ref
            criteria_tmd.update_term(
                *self.barcode.min_term(term_above=reference, term_below=maximum_target)
            )
            criteria_tmd = self.barcode.curate_stop_criterion(criteria_tmd, criteria_tmd)

        current_section.stop_criteria["TMD"] = criteria_tmd

        return current_section.next()


class TMDApicalAlgo(TMDAlgo):
    """TreeGrower of TMD apical growth."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._found_last_bif = False

    def initialize(self):
        """Initialize the algorithm.

        TMD basic grower of an apical tree based on path distance.
        Initializes the tree grower and computes the apical distance using the input barcode.
        """
        # pylint: disable=C0415
        from tmd.Topology.analysis import find_apical_point_distance_smoothed as ap_dist

        stop, num_sec = super().initialize()
        if self.params.get("has_apical_tuft", False):
            self.apical_point_distance_from_soma = ap_dist(self.ph_angles)
        else:
            # If cell does not have a tuft return the distance before last bifurcation
            # this will result in a point very close to the proximal apical point.
            step_size = self.params["step_size"]["norm"]["mean"]
            selected_length = list(self.barcode.bifs.values())[-1]
            # From last bifurcation subtract 10 step sizes for apical point distance
            self.apical_point_distance_from_soma = selected_length - 10 * step_size
        return stop, num_sec

    def bifurcate(self, current_section):
        """When the section bifurcates two new sections need to be created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.

        Returns:
            tuple[dict, dict]: Two dictionaries containing the data of the two children sections.
        """
        self.barcode.remove_bif(current_section.stop_criteria["TMD"].bif_id)
        ang = self.barcode.angles[current_section.stop_criteria["TMD"].bif_id]
        current_pd = self.metric(current_section)
        first_point = np.array(current_section.last_point)

        if current_section.process == "major":
            dir1, dir2 = bif_methods["directional"](current_section.direction, angles=ang)

            if not self._found_last_bif:
                self.apical_section = current_section.id

            if current_pd <= self.apical_point_distance_from_soma:
                process1 = "major"
                process2 = "secondary"
            else:
                process1 = "secondary"
                process2 = "secondary"

                if not self._found_last_bif:
                    self._found_last_bif = True
        else:
            dir1, dir2 = self.bif_method(current_section.history(), angles=ang)
            process1 = "secondary"
            process2 = "secondary"

            if not self._found_last_bif:
                self.apical_section = current_section.id

                if current_pd > self.apical_point_distance_from_soma:
                    self._found_last_bif = True

        stop1, stop2 = self.get_stop_criteria(current_section)

        return (
            section_data(dir1, first_point, stop1, process1),
            section_data(dir2, first_point, stop2, process2),
        )


class TMDGradientAlgo(TMDApicalAlgo):
    """TreeGrower of TMD apical growth."""

    def _majorize_process(self, section, stop, process, input_dir):
        """Currates the non-major processes to apply a gradient to large components."""
        bias_length = self.params["bias_length"] * self.persistence_length
        difference = stop.expected_maximum_length()
        if difference > bias_length:
            direction1 = (1.0 - self.params["bias"]) * np.array(input_dir)
            direction2 = self.params["bias"] * np.array(section.direction)
            direct = np.add(direction1, direction2)
            return "major", direct / norm(direct)
        return process, input_dir

    def bifurcate(self, current_section):
        """When the section bifurcates two new sections need to be created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.

        Returns:
            tuple[dict, dict]: Two dictionaries containing the two children sections data.
        """
        s1, s2 = super().bifurcate(current_section)

        if s1["process"] != "major":
            s1["process"], s1["direction"] = self._majorize_process(
                current_section, s1["stop"]["TMD"], s1["process"], s1["direction"]
            )
        if s2["process"] != "major":
            s2["process"], s2["direction"] = self._majorize_process(
                current_section, s2["stop"]["TMD"], s2["process"], s2["direction"]
            )
        return s1, s2



"""Astrocyte section growers."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from collections import namedtuple

import numpy as np

from neurots.generate.section import SectionGrowerPath
from neurots.morphmath.utils import from_to_direction
from neurots.morphmath.utils import get_random_point
from neurots.morphmath.utils import in_squared_proximity
from neurots.morphmath.utils import normalize_inplace

L = logging.getLogger(__name__)


NextPointData = namedtuple("NextPointData", ["point", "direction", "segment_length"])


def grow_to_target(start_point, start_direction, target_point, segment_length, p=0.5):
    """Grow towards the target_point with segment_length step from the given point and direction.

    Args:
        start_point (numpy.ndarray): Starting point of the grower.
        start_direction (numpy.ndarray): Normalized initial direction.
        target_point (numpy.ndarray): Target point to grow to.
        segment_length (numpy.ndarray): The step size of the grower.
        p (float, optional): Influence from the target.
            If zero the grower will grow a straight line
            from the start point along the initial direction, and if 1.0 it will grow
            a straight line from the start point to the target point. Defaults to 0.5

    Returns:
        list[numpy.ndarray]: A list of the generated 3D points.
    """
    target_proximity = (1.5 * segment_length) ** 2

    point = start_point.copy()
    direction = start_direction.copy()

    points = []
    while not in_squared_proximity(point, target_point, target_proximity):

        target_direction = from_to_direction(point, target_point)
        direction = (1.0 - p) * direction + p * target_direction

        # zeros direction results from an initial direction which opposite to the target one
        # and the p = 0.5 . In that case the target_direction is used instead.
        direction = (
            target_direction if np.allclose(direction, 0.0) else normalize_inplace(direction)
        )

        point += segment_length * direction
        points.append(point)

    # add the target point if the new point does not coincide
    if not np.allclose(point, target_point):
        points.append(target_point)

    return points


class SectionSpatialGrower(SectionGrowerPath):
    """Section grower that is influenced by a point cloud.

    .. note::
        The main difference with a regular grower is that randomness is calculated by the nearest
        neighbor in the point cloud. This way there is a local influence by the distribution of the
        seeds. As the seeds get removed when a new point is created, the new points are influenced
        only by the seeds that area available. This leads to a more spread out occupancy of space.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._influence_distance = self.context.influence_distance(self.step_size_distribution.loc)

    @property
    def point_cloud(self):
        """Returns point cloud from context."""
        return self.context.point_cloud

    @property
    def morphology_points(self):
        """Returns entire morphology's points from context."""
        return self.context.morphology_points

    def _add_new_data(self, point, direction, segment_length):
        """Update relevant data after a next point action.

        Following a next point action, make an update of the relevant data and remove a hemisphere
        of seed points around center point, directed backwards to direction.
        """
        self.update_pathlength(segment_length)

        self.points.append(point)
        self.morphology_points.append(point)
        self.latest_directions.append(direction)

        kill_distance = self.context.kill_distance(segment_length)
        self.point_cloud.remove_hemisphere(point, -direction, kill_distance)

        return NextPointData(point, direction, segment_length)

    def first_point(self):
        """Generates the first point of the section.

        The generation depends on the growth method and the previous point. This gives
        flexibility to implement a specific computation for the first point, and ensures the
        section has at least one point.

        .. warning::
            The first point should be always called from the initialization when the section grower
            is added in the active list. It guarantees at least two point sections.
        """
        new_direction = normalize_inplace(0.8 * self.direction + 0.2 * self.history())
        segment_length = self.step_size_distribution.draw_positive()
        new_point = self.last_point + segment_length * new_direction
        self._add_new_data(new_point, new_direction, segment_length)

    def _neighbor_contribution(self, current_point):
        """Return the direction considering the nearest neighbor.

        Random contribution has a two case component: the influence from the closest
        seed attractor if any, otherwise a random direction.
        """
        pcloud_direction = self.point_cloud.nearest_neighbor_direction(
            current_point, self._influence_distance
        )

        if pcloud_direction is not None:
            return pcloud_direction

        return get_random_point(random_generator=self._rng)

    def next_direction(self, current_point):
        """Return the next direction.

        Given a starting point, find the new direction taking into account
        three contributions: the targeting, history and random contributions. Space
        colonization random contribution is calculated by the closest seed point
        in the point cloud.

        Args:
            current_point (numpy.ndarray): The current point.

        Returns:
            numpy.ndarray: The normalized direction.
        """
        return normalize_inplace(
            self.params.targeting * self.direction
            + self.params.history * self.history()
            + self.params.randomness * self._neighbor_contribution(current_point)
        )

    def _next_point(self, current_point):
        """Create the next point.

        Also update if no collision takes place or if the section type is endfoot.
        In addition, if the section type is endfoot, grow all points until
        the endfoot target and return None, to terminate.
        """
        segment_length = self.step_size_distribution.draw_positive()

        new_direction = self.next_direction(current_point)
        new_point = current_point + segment_length * new_direction

        if self.process == "endfoot":
            self._add_new_data(new_point, new_direction, segment_length)
            self._grow_endfoot_section(new_point, new_direction, segment_length)
            return None

        if self.context.collision_handle(new_point, segment_length):
            return None

        return self._add_new_data(new_point, new_direction, segment_length)

    def _grow_endfoot_section(self, initial_point, initial_direction, segment_length):
        """Creates all the points of the endfoot section at once and makes the relevant updates."""
        target_point = self.context.endfeet_targets.points[self.stop_criteria["target_id"]]

        grown_points = grow_to_target(
            initial_point, initial_direction, target_point, segment_length
        )

        for p in grown_points:
            self.points.append(p)
            self.morphology_points.append(p)
            self.point_cloud.remove_points_around(p, segment_length)

    def next(self):
        """Creates one point and returns the next state: bifurcate, terminate or continue."""
        if not self._next_point(self.last_point):
            self.children = 0
            return "terminate"

        if self.check_stop():
            return "continue"

        if self.children == 0:
            return "terminate"

        return "bifurcate"



"""Test neurots.morphmath.utils code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import numpy as np
from numpy import testing as npt

from neurots.morphmath import utils as tested


def test_norm():
    vector = np.array([0.24428195, 0.84795862, 0.70495863])
    npt.assert_allclose(tested.norm(vector), np.linalg.norm(vector))

    vector = np.zeros(3)
    npt.assert_allclose(tested.norm(vector), 0.0)


def test_from_to_direction():

    point = np.array([0.51374375, 0.39753749, 0.27568173])
    target = np.array([0.5932438, 0.96379423, 0.49277981])

    expected_length = 0.6116359455469541
    expected_direction = np.array([0.12997936, 0.92580684, 0.35494657])

    direction = tested.from_to_direction(point, target, return_length=False)
    npt.assert_allclose(direction, expected_direction)

    direction, length = tested.from_to_direction(point, target, return_length=True)
    npt.assert_allclose(direction, expected_direction)
    npt.assert_allclose(length, expected_length)


def test_in_same_halfspace():

    points = np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0], [2.0, 2.0, 2.0]])

    ref_point = np.array([0.5, 0.5, 0.5])

    direction = -np.array([1.0, 1.0, 1.0])
    direction /= np.linalg.norm(direction)

    mask = tested.in_same_halfspace(points - ref_point, direction)
    npt.assert_array_equal(mask, [True, False, False])

    mask, dots = tested.in_same_halfspace(points - ref_point, direction, return_dots=True)
    npt.assert_array_equal(mask, [True, False, False])

    npt.assert_allclose(dots, -np.dot(ref_point, direction))


def test_in_squared_proximity():
    point1 = np.array([0.0, 0.0, 1.0])
    point2 = np.array([0.0, 0.0, 3.0])

    squared_proximity = 4.0
    assert tested.in_squared_proximity(point1, point2, squared_proximity)

    squared_proximity = 3.9
    assert not tested.in_squared_proximity(point1, point2, squared_proximity)


def test_ball_query():

    points = np.array(
        [
            [0.0, 0.0, -2.0],
            [0.0, 0.0, -1.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 2.0],
        ]
    )

    ids = tested.ball_query(points, np.array([0.0, 0.0, 0.0]), 1.0)
    npt.assert_array_equal(ids, [1, 2, 3])


def test_upper_half_ball_query():

    points = np.array([[10.0, 10.0, 10.0], [11.0, 11.0, 11.0]])
    ids = tested.upper_half_ball_query(points, np.zeros(3), 0.1, np.array([1.0, 0.0, 0.0]))
    assert ids.size == 0

    points = np.array(
        [
            [0.0, 0.0, -2.0],
            [0.0, 0.0, -1.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 2.0],
        ]
    )

    ids = tested.upper_half_ball_query(
        points, np.array([0.0, 0.0, 0.0]), 1.0, np.array([-1.0, 0.0, -1.0])
    )
    npt.assert_array_equal(ids, [1])



"""Test neurots.astrocyte.space_colonization code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import itertools

import numpy as np
from mock import Mock
from mock import patch
from numpy import testing as npt

from neurots.astrocyte import space_colonization as tested

PCLOUD_POINTS = np.array(
    [
        [0.0, 0.0, 0.0],
        [0.1, 0.1, 0.1],
        [0.2, 0.2, 0.2],
        [0.3, 0.3, 0.3],
        [0.5, 0.5, 0.5],
        [0.6, 0.6, 0.6],
    ]
)


def _point_cloud():

    point_cloud = Mock(points=PCLOUD_POINTS, remove_hemisphere=Mock())

    values = [
        np.array([0.0, 0.0, 1.0]),
        np.array([0.0, 1.0, 0.0]),
        np.array([1.0, 0.0, 0.0]),
    ]

    point_cloud.nearest_neighbor_direction = Mock(side_effect=itertools.cycle(values))

    ids = [np.array([0, 1, 2]), np.array([1, 2])]

    point_cloud.ball_query = Mock(side_effect=itertools.cycle(ids))

    return point_cloud


def test_majorize():

    process = "secondary"

    stop = Mock()
    stop.expected_maximum_length.return_value = np.inf

    target_distance = None

    result = tested._majorize_process(process, stop, target_distance)
    assert result == "secondary"

    stop.expected_maximum_length.return_value = 100.0
    target_distance = 90.0

    result = tested._majorize_process(process, stop, target_distance)
    assert result == "major"

    target_distance = 120.0
    result = tested._majorize_process(process, stop, target_distance)
    assert result == "secondary"


def test_repulsion():

    point = np.zeros(3)

    points = np.array([[0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]])

    vectors = points - point

    result = tested._repulsion(points[:1], point, 1.0)
    expected = vectors[0] * np.exp(-1)
    npt.assert_allclose(result, expected)

    result = tested._repulsion(points[:1], point, 0.5)
    expected = vectors[0] * np.exp(-2.0)
    npt.assert_allclose(result, expected)

    result = tested._repulsion(points[1:2], point, 1.0)
    expected = 0.5 * vectors[1] * np.exp(-1.0 * 2.0)
    npt.assert_allclose(result, expected)

    result = tested._repulsion(points[1:2], point, 0.5)
    expected = 0.5 * vectors[1] * np.exp(-2.0 * 2.0)
    npt.assert_allclose(result, expected)

    result = tested._repulsion(points, point, 0.5)
    expected = (
        vectors[0] * np.exp(-2.0)
        + vectors[1] * np.exp(-2.0 * 2.0) / 2.0
        + vectors[2] * np.exp(-2.0 * 3.0) / 3.0
    ) / 3.0
    npt.assert_allclose(result, expected)


def test_fallback_strategy():

    repulsion = np.array([0.0, 1.0, 2.0])

    mock_dir1 = np.array([1.0, 0.0, 0.0])
    mock_dir2 = np.array([0.0, 1.0, 0.0])

    with patch(
        "neurots.astrocyte.space_colonization.directional",
        return_value=(mock_dir1.copy(), mock_dir2.copy()),
    ):

        section_direction = np.array([0.0, 0.0, 1.0])

        dir1, dir2 = tested._fallback_strategy(section_direction, None, repulsion)

        expected_dir1 = section_direction - repulsion
        expected_dir1 /= np.linalg.norm(expected_dir1)

        npt.assert_allclose(dir1, expected_dir1)

        expected_dir2 = mock_dir2 - repulsion
        expected_dir2 /= np.linalg.norm(expected_dir2)
        npt.assert_allclose(dir2, expected_dir2)


def test_colonization_strategy_secondary():

    vectors = np.array(
        [
            [0.14525965, 0.18989378, 0.73957937],
            [0.72947892, 0.28593541, 0.23994719],
            [0.51646674, 0.20958965, 0.25134127],
        ]
    )

    repulsion = np.array([0.0, 1.0, 2.0])

    dir1, dir2 = tested._colonization_strategy_secondary(vectors, repulsion)

    vecs = vectors - repulsion
    lengths = np.linalg.norm(vecs, axis=1)

    smallest_length_id = np.argmin(lengths)
    expected_dir1 = vecs[smallest_length_id] / lengths[smallest_length_id]
    npt.assert_allclose(dir1, expected_dir1)

    uvecs = vecs / lengths[:, None]

    largest_angle_id = np.argmin(uvecs.dot(dir1))
    expected_dir2 = uvecs[largest_angle_id]

    npt.assert_allclose(dir2, expected_dir2)


def test_colonization_strategy_primary():

    vectors = np.array(
        [
            [0.14525965, 0.18989378, 0.73957937],
            [0.72947892, 0.28593541, 0.23994719],
            [0.51646674, 0.20958965, 0.25134127],
        ]
    )

    repulsion = np.array([0.0, 1.0, 2.0])
    direction = np.array([0.0, 0.0, 1.0])

    dir1, dir2 = tested._colonization_strategy_primary(direction, vectors, repulsion)

    vecs = vectors - repulsion
    lengths = np.linalg.norm(vecs, axis=1)

    expected_dir1 = direction - repulsion
    expected_dir1 /= np.linalg.norm(expected_dir1)

    npt.assert_allclose(dir1, expected_dir1)

    uvecs = vecs / lengths[:, None]

    largest_angle_id = np.argmin(uvecs.dot(dir1))
    expected_dir2 = uvecs[largest_angle_id]

    npt.assert_allclose(dir2, expected_dir2)


def _section(process_type):

    mock = Mock()
    mock.direction = np.array([1.0, 0.0, 0.0])
    mock.last_point = np.array([0.5, 0.5, 0.5])
    mock.segment_length = 0.1
    mock.kill_distance = 1.0
    mock.influence_distance = 2.0

    mock.process = process_type

    mock.step_size_distribution = Mock(params={"mean": 0.1, "std": 0.001})

    mock.stop_criteria = {"max_target_distance": 1.0, "TMD": None, "target_id": 3}
    return mock


def test_colonization_split():

    section = _section("major")
    parameters = {"step_size": {"norm": {"mean": 1.0}}}

    context = Mock(
        point_cloud=Mock(points=np.array([[0.2, 0.3, 0.4], [0.5, 0.6, 0.7]])),
        morphology_points=Mock(data=np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])),
    )
    point_cloud = context.point_cloud

    module = "neurots.astrocyte.space_colonization."

    with patch(module + "_repulsion") as repulsion, patch(
        module + "upper_half_ball_query"
    ) as half_ball_query, patch(module + "_fallback_strategy") as fallback_strategy, patch(
        module + "_colonization_strategy_primary"
    ) as primary_strategy, patch(
        module + "_colonization_strategy_secondary"
    ) as secondary_strategy:

        repulsion.return_value = np.array([3.0, 2.0, 1.0])
        half_ball_query.return_value = []

        # not enough seed points will trigger the fallback strategy
        point_cloud.partial_ball_query.return_value = np.array([0], dtype=int)
        point_cloud.upper_half_ball_query.return_value = np.array([0], dtype=int)

        fallback_strategy.return_value = (
            np.array([0.0, 0.0, 1.0]),
            np.array([1.0, 0.0, 0.0]),
        )

        # for both major
        section.process = "major"
        dir1, typ1, dir2, typ2 = tested._colonization_split(section, None, parameters, context)

        npt.assert_allclose(dir1, [0, 0, 1])
        npt.assert_allclose(dir2, [1, 0, 0])

        assert typ1 == "major"
        assert typ2 == "secondary"

        # and secondary process types
        section.process = "secondary"
        dir1, typ1, dir2, typ2 = tested._colonization_split(section, None, parameters, context)

        npt.assert_allclose(dir1, [0, 0, 1])
        npt.assert_allclose(dir2, [1, 0, 0])

        assert typ1 == "secondary"
        assert typ2 == "secondary"

        # 2 or more point ids and a major section will not trigger the fallback
        point_cloud.partial_ball_query.return_value = np.array([0, 1], dtype=int)
        point_cloud.upper_half_ball_query.return_value = np.array([0, 1], dtype=int)

        # major triggers primary strategy
        section.process = "major"
        primary_strategy.return_value = (
            np.array([0.0, 1.0, 0.0]),
            np.array([-1.0, 0.0, 0.0]),
        )

        dir1, typ1, dir2, typ2 = tested._colonization_split(section, None, parameters, context)

        npt.assert_allclose(dir1, [0, 1, 0])
        npt.assert_allclose(dir2, [-1, 0, 0])

        assert typ1 == "major"
        assert typ2 == "secondary"

        # secondary type triggers secondary strategy
        section.process = "secondary"
        secondary_strategy.return_value = (
            np.array([0.0, 0.0, 1.0]),
            np.array([0.0, 0.0, -1.0]),
        )

        dir1, typ1, dir2, typ2 = tested._colonization_split(section, None, parameters, context)

        npt.assert_allclose(dir1, [0, 0, 1])
        npt.assert_allclose(dir2, [0, 0, -1])

        assert typ1 == "secondary"
        assert typ2 == "secondary"

        # finally if for any reason the strategies return identical directions
        # the dirs are recalculated using the fallback strategy
        secondary_strategy.return_value = (
            np.array([0.0, 0.0, 1.0]),
            np.array([0.0, 0.0, 1.0]),
        )

        dir1, typ1, dir2, typ2 = tested._colonization_split(section, None, parameters, context)

        npt.assert_allclose(dir1, [0, 0, 1])
        npt.assert_allclose(dir2, [1, 0, 0])

        assert typ1 == "secondary"
        assert typ2 == "secondary"


def test_add_attraction_bias():

    target_point = np.array([0.0, 0.0, 100.0])
    current_point = np.array([0.0, 0.0, 0.0])

    def attraction_function(x):
        return 1.0 - x

    max_target_distance = 100.0

    direction = np.array([1.0, 0.0, 0.0])

    new_direction = tested._add_attraction_bias(
        current_point, target_point, max_target_distance, direction, attraction_function
    )

    npt.assert_allclose(new_direction, direction)

    current_point = np.array([0.0, 0.0, 99.99999])

    new_direction = tested._add_attraction_bias(
        current_point, target_point, max_target_distance, direction, attraction_function
    )

    # super close to target -> direction to target only
    npt.assert_allclose(new_direction, [0.0, 0.0, 1.0], atol=1e-6)


def test_colonization_split_with_target_influence():

    parameters = {
        "bias": 1.0,
        "distance_soma_target": 1.0,
        "step_size": {"norm": {"mean": 1.0}},
    }

    section = _section("major")
    section.target = Mock(available=True, xyz=np.array([2.0, 2.0, 2.0]))

    endfeet = Mock()
    endfeet.active = np.ones(5, dtype=bool)
    endfeet.points = np.random.random((5, 3))

    point_cloud = None
    context = Mock(endfeet_targets=endfeet, point_cloud=point_cloud)

    module = "neurots.astrocyte.space_colonization."

    with patch(module + "_colonization_split") as mock_colonization_split, patch(
        module + "in_squared_proximity"
    ) as mock_in_squared_proximity, patch(
        module + "_add_attraction_bias"
    ) as mock_add_attraction_bias, patch(
        module + "_majorize_process"
    ) as mock_majorize_process:

        mock_colonization_split.return_value = (
            np.array([0.0, 0.0, 1.0]),
            None,
            np.array([1.0, 0.0, 0.0]),
            None,
        )

        target_id = section.stop_criteria["target_id"]
        # if there are not active endfeet the space colonization directions are returned
        endfeet.active[target_id] = False

        dir1, typ1, dir2, typ2 = tested._colonization_split_with_target_influence(
            section, None, parameters, context
        )

        npt.assert_allclose(dir1, [0.0, 0.0, 1.0])
        npt.assert_allclose(dir2, [1.0, 0.0, 0.0])

        assert typ1 == "major"
        assert typ2 == "secondary"

        endfeet.active[target_id] = True

        # if point in proximity of target change process type to endfoot
        # the dir2 will be the direction to the target
        mock_in_squared_proximity.return_value = True

        dir1, typ1, dir2, typ2 = tested._colonization_split_with_target_influence(
            section, None, parameters, context
        )

        npt.assert_allclose(dir1, [0.0, 0.0, 1.0])

        expected_direction = endfeet.points[target_id] - section.last_point
        expected_direction /= np.linalg.norm(expected_direction)
        npt.assert_allclose(dir2, expected_direction)

        assert typ1 == "major"
        assert typ2 == "endfoot"

        # target is deactivated when an endfoot is assigned to it
        assert not endfeet.active[target_id]

        # if point not in proximity of the target and the section is
        # of major type, then attraction bias is added
        mock_in_squared_proximity.return_value = False
        section.process = "major"
        endfeet.active[target_id] = True

        mock_add_attraction_bias.return_value = np.array([-1.0, 0.0, 0.0])

        dir1, typ1, dir2, typ2 = tested._colonization_split_with_target_influence(
            section, None, parameters, context
        )

        npt.assert_allclose(dir1, [-1.0, 0.0, 0.0])
        npt.assert_allclose(dir2, [1.0, 0.0, 0.0])

        assert typ1 == "major"
        assert typ2 == "secondary"

        # if point not in proximity of the target and the section is
        # of not major type, then the process type is majorized

        section.process = "secondary"
        mock_majorize_process.return_value = "major"

        dir1, typ1, dir2, typ2 = tested._colonization_split_with_target_influence(
            section, None, parameters, context
        )

        npt.assert_allclose(dir1, [0.0, 0.0, 1.0])
        npt.assert_allclose(dir2, [1.0, 0.0, 0.0])

        assert typ1 == "major"
        assert typ2 == "secondary"



"""Point array class."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


class DynamicPointArray:
    """Store points in a numpy array and automatically resizes when its capacity is reached.

    It is used by algorithms that require the points as a :class:`numpy.array` and append points
    incrementally.

    Args:
        initial_capacity (int): The initial capacity of the array.
        resize_factor (float): The factor used to increase the capacity of the array.
    """

    def __init__(self, initial_capacity=100000, resize_factor=2.0):

        self._size = 0
        self._capacity = initial_capacity
        self._resize_factor = resize_factor
        self._data = np.empty((initial_capacity, 3), dtype=np.float32)

    def __len__(self):
        """Return the length of the array."""
        return self._size

    @property
    def capacity(self):
        """Returns the current capacity of the array."""
        return self._capacity

    @property
    def data(self):
        """Return filled data."""
        return self._data[: self._size]

    def _resize_capacity(self):
        """Resizes the capacity when the size equals capacity."""
        self._capacity = int(self._resize_factor * self._capacity)
        self._data = np.resize(self._data, (self._capacity, 3))

    def append(self, point):
        """Append a point to the array."""
        if self._size == self._capacity:
            self._resize_capacity()

        self._data[self._size] = point
        self._size += 1



"""Test neurots.generate.section code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from neurots.generate import section
from neurots.generate.algorithms.common import TMDStop
from neurots.morphmath import sample

EXPECTED_WEIGHTS = np.array([0.01831564, 0.04978707, 0.13533528, 0.36787944, 1.0])


@pytest.fixture
def SEG_LEN():
    return sample.Distr({"norm": {"mean": 1.0, "std": 0.2}})


def test_MEMORY_WEIGHTS():
    assert section.MEMORY == 5
    assert_array_almost_equal(section.WEIGHTS, EXPECTED_WEIGHTS)


def test_SectionGrower(SEG_LEN):
    # Create basic section for testing purposes
    s = section.SectionGrower(
        None, None, [0.0, 0.0, 0.0], [0.0, 1.0, 0.0], 0.0, 0.0, None, None, SEG_LEN, 0.0
    )
    # Test trivial history is zeros.
    assert_array_almost_equal(s.history(), np.zeros(3))
    # Test history of one point.
    s.points.append([0.0, 1.0, 0.0])
    s.latest_directions.append([0.0, 1.0, 0.0])
    assert_array_almost_equal(s.history(), np.array([0.0, 1.0, 0.0]))
    # Test history of two point.
    s.points.append([0.0, 1.0, 1.0])
    s.latest_directions.append([0.0, 0.0, 1.0])
    assert_array_almost_equal(s.history(), np.array([0.0, 0.34525776, 0.9385079]))
    s.points.append([0.0, 1.0, 2.0])
    s.latest_directions.append([0.0, 0.0, 1.0])
    assert_array_almost_equal(s.history(), np.array([0.0, 0.0984573, 0.99514128]))
    s.points.append([0.0, 1.0, 3.0])
    s.latest_directions.append([0.0, 0.0, 1.0])
    assert_array_almost_equal(s.history(), np.array([0.0, 0.03310225, 0.99945197]))
    s.points.append([1.0, 2.0, 4.0])
    s.latest_directions.append([1.0, 1.0, 1.0] / np.linalg.norm([1.0, 1.0, 1.0]))
    assert_array_almost_equal(s.history(), np.array([0.41177931, 0.42484243, 0.80619272]))

    tmd_stop = TMDStop(1, 26.3027, 0, 633.5966, 40.0)
    # Check small section
    s = section.SectionGrowerExponentialProba(
        None,
        None,
        [0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        0.0,
        0.0,
        tmd_stop,
        None,
        SEG_LEN,
        0.0,
    )
    # Check that the growth will continue for single point section
    assert s.check_stop()
    with pytest.raises(NotImplementedError):
        s.get_val()


def test_history(SEG_LEN):
    s = section.SectionGrower(
        None, None, [0.0, 0.0, 0.0], [0.0, 1.0, 0.0], 0.0, 0.0, None, None, SEG_LEN, 0.0
    )
    assert_array_almost_equal(s.history(), np.array([0.0, 0.0, 0.0]))
    s.latest_directions = np.array([[0.0, 0.0, 0.0]])
    assert_array_almost_equal(s.history(), np.array([0.0, 0.0, 0.0]))
    s.latest_directions = np.array([[0.0, 1.0, 0.0]])
    assert_array_almost_equal(s.history(), np.array([0.0, 1.0, 0.0]))
    s.latest_directions = np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    assert_array_almost_equal(s.history(), np.array([0.0, 0.34525776, 0.9385079]))
    s.latest_directions = np.array([[0.0, 0.0, 0.00000001]])
    assert_array_almost_equal(s.history(), np.array([0.0e00, 0.0e00, 1.0e-08]))



"""Define a set of distributions from which to sample.

Currently the available sampling features are:
1. soma_size
2. n_neurites
3. trunk_orientation_deviations
"""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from distributions import sample

distributions_exemplar = {
    "soma_size": {"norm": {"mean": 8, "std": 1.18}},
    "n_neurites": {"norm": {"mean": 6.75, "std": 1.37}},
    "trunk_orientation_deviation": {
        "data": {
            "bins": [
                -1.0017981,
                -0.93768791,
                -0.87357773,
                -0.80946755,
                -0.74535737,
                -0.68124718,
                -0.617137,
                -0.55302682,
                -0.48891663,
                -0.42480645,
                -0.36069627,
                -0.29658609,
                -0.2324759,
                -0.16836572,
                -0.10425554,
                -0.04014535,
                0.02396483,
                0.08807501,
                0.1521852,
                0.21629538,
                0.28040556,
                0.34451574,
                0.40862593,
                0.47273611,
                0.53684629,
                0.60095648,
                0.66506666,
                0.72917684,
                0.79328702,
                0.85739721,
                0.92150739,
                0.98561757,
                1.04972776,
                1.11383794,
                1.17794812,
                1.2420583,
                1.30616849,
                1.37027867,
                1.43438885,
                1.49849904,
            ],
            "weights": [
                3.0,
                6.0,
                6.0,
                4.0,
                3.0,
                5.0,
                4.0,
                3.0,
                4.0,
                6.0,
                2.0,
                1.0,
                4.0,
                0.0,
                6.0,
                0.0,
                1.0,
                9.0,
                1.0,
                2.0,
                1.0,
                4.0,
                2.0,
                2.0,
                3.0,
                2.0,
                2.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                3.0,
                1.0,
                1.0,
                3.0,
                0.0,
                3.0,
                1.0,
                2.0,
            ],
        }
    },
}

# import sampler

soma_radius = sample.soma_size(distributions_exemplar)
n_neurites = sample.n_neurites(distributions_exemplar)
trunk_angles = sample.trunk_angles(distributions_exemplar, n_neurites)



"""Test the examples."""

# Copyright (C) 2022  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import importlib.util as ilu
from pathlib import Path

EXAMPLES = Path(__file__).parent.parent / "examples"
DATA = EXAMPLES / "data"


def _run_example(example, tmpdir, data):
    name = example.split("/")[-1]
    spec = ilu.spec_from_file_location(name, example)
    current_example = ilu.module_from_spec(spec)
    spec.loader.exec_module(current_example)

    current_example.run(tmpdir, data)


def test_extract_synthesis_inputs(tmpdir):
    _run_example("examples/extract_synthesis_inputs.py", tmpdir, DATA)
    assert sorted([i.relto(tmpdir) for i in tmpdir.listdir()]) == [
        "generated_cell.asc",
        "generated_cell.h5",
        "generated_cell.swc",
        "test_distr.json",
        "test_params.json",
    ]


def test_synthesize_neuron_external_diameter(tmpdir):
    _run_example("examples/synthesize_neuron_external_diameter.py", tmpdir, DATA)
    assert sorted([i.relto(tmpdir) for i in tmpdir.listdir()]) == [
        "generated_cell.asc",
        "generated_cell.h5",
        "generated_cell.swc",
    ]


def test_synthesize_neuron_with_diameters(tmpdir):
    _run_example("examples/synthesize_neuron_with_diameters.py", tmpdir, DATA)
    assert sorted([i.relto(tmpdir) for i in tmpdir.listdir()]) == [
        "generated_cell.asc",
        "generated_cell.h5",
        "generated_cell.swc",
    ]


def test_synthesize_neurons(tmpdir):
    _run_example("examples/synthesize_neurons.py", tmpdir, DATA)
    assert sorted([i.relto(tmpdir) for i in tmpdir.listdir()]) == [
        f"generated_cell_{i}.swc" for i in range(10)
    ]


def test_synthesize_single_neuron(tmpdir):
    _run_example("examples/synthesize_single_neuron.py", tmpdir, DATA)
    assert sorted([i.relto(tmpdir) for i in tmpdir.listdir()]) == [
        "generated_cell.asc",
        "generated_cell.h5",
        "generated_cell.swc",
    ]



"""NeuroTS class: Tree."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy

from neurots.astrocyte.section import SectionSpatialGrower
from neurots.astrocyte.space_colonization import SpaceColonization
from neurots.astrocyte.space_colonization import SpaceColonizationTarget
from neurots.generate.tree import TreeGrower

GROWTH_ALGORITHMS = {
    "tmd_space_colonization": SpaceColonization,
    "tmd_space_colonization_target": SpaceColonizationTarget,
}

SECTION_GROWERS = {
    "tmd_space_colonization": SectionSpatialGrower,
    "tmd_space_colonization_target": SectionSpatialGrower,
}


class TreeGrowerSpaceColonization(TreeGrower):
    """Tree class."""

    def _initialize_algorithm(self):
        """Initialization steps."""
        grow_meth = GROWTH_ALGORITHMS[self.params["growth_method"]]

        growth_algo = grow_meth(
            input_data=self.distr,
            params=self.params,
            start_point=self.point,
            context=self.context,
            skip_validation=self.skip_validation,
            random_generator=self._rng,
        )

        stop, num_sec = growth_algo.initialize()

        section = self.add_section(
            parent=None,
            direction=self.direction,
            first_point=self.point,
            stop=copy.deepcopy(stop),
            process="major",
            pathlength=0.0,
            children=2 if num_sec > 1 else 0,
        )

        # First section of the tree has at least one point.
        section.first_point()

        return growth_algo

    def add_section(
        self, parent, direction, first_point, stop, pathlength, process=None, children=0
    ):
        """Generates a section from the parent section "act" from all the required information.

        The section is added to the neuron.sections and activated.
        """
        SGrower = SECTION_GROWERS[self.params["growth_method"]]

        sec_grower = SGrower(
            parent=parent,
            children=children,
            first_point=first_point,
            direction=direction,
            parameters=self._section_parameters,
            process=process,
            stop_criteria=copy.deepcopy(stop),
            step_size_distribution=self.seg_length_distr,
            pathlength=pathlength,
            context=self.context,
            random_generator=self._rng,
        )

        self.active_sections.append(sec_grower)
        return sec_grower



"""Class to collect all TMD related info used in NeuroTS."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
from collections import OrderedDict

import numpy as np

from neurots.basic import round_num
from neurots.utils import NeuroTSError


class Barcode:
    """Class to generate the barcode structure.

    The barcode structure is essential for the TMD based growth algorithms.

    Args:
        ph_angles (list of lists): list of barcodes and angles of 6 elements:

            * either::

                [
                    end_point,
                    start_point,
                    4D_angles
                ]

            * or equivalent::

                [
                    Termination,
                    Bifurcation,
                    Angle parent - child (x-y),
                    Angle parent - child (z),
                    Angle child1 - child2 (x-y),
                    Angle child1 - child2 (z)
                ]

    Returns:
       The ph_angles will be decomposed in the following dictionaries::

           {
               angles: {ID: 4D_angles}
               bifs: {ID: start_point}
               terms: {ID, end_point}
           }
    """

    def __init__(self, ph_angles):
        """Initialize the ph_angles into a barcode object."""
        # Sort persistence bars according to bifurcation
        ph_angles.sort(key=lambda x: x[1])

        self.bifs = OrderedDict()
        self.terms = OrderedDict()
        self.angles = OrderedDict()

        for bar_index, tmd_bar in enumerate(ph_angles):
            self.bifs[bar_index] = round_num(tmd_bar[1])
            self.terms[bar_index] = round_num(tmd_bar[0])
            self.angles[bar_index] = tmd_bar[2:]

        # Sort the terminations according to value to optimize access
        self.terms = OrderedDict(sorted(self.terms.items(), key=lambda x: x[1]))
        # Bifurcation at 0 is trivial so it should be removed
        del self.bifs[0]

    @staticmethod
    def validate_persistence(ph_angles):
        """Checks if data are in the expected format.

        The input barcodes should follow the rules:
        1. Bar: (start, end), end > start
        """
        for p in ph_angles:
            if p[0] <= p[1]:
                return False
        return True

    def get_bar(self, bar_id):
        """Returns the pair of (bifurcation, termination) that corresponds to the input index.

        The trivial bifurcation of id=0 is not included.
        """
        if bar_id == 0:
            return (0, self.terms[bar_id])
        return (self.bifs[bar_id], self.terms[bar_id])

    def get_persistence_length(self):
        """Returns the maximum bar length."""
        return self.terms[0]

    def remove_bif(self, bar_id):
        """Remove a bifurcation that has been used if bif_id is not None."""
        if bar_id is not None:
            del self.bifs[bar_id]

    def remove_term(self, bar_id):
        """Remove a termination that has been used, if term_id is not None."""
        if bar_id is not None:
            del self.terms[bar_id]

    def get_term(self, bar_id):
        """Returns a termination based on index if the input ID exists of infinity.

        If it doesn't exist the branch will terminate
        as it gets term = -infinity.
        """
        try:
            return self.terms[bar_id]
        except KeyError:
            # bar ID does not exist
            return -np.inf

    def get_term_between(self, bar_id, above=0.0, below=np.inf):
        """Returns a termination based on index.

        If the index exists and its value is between the above / below thresholds, the termination
        is returned.
        If it doesn't exist the branch will terminate as it gets `term = -infinity`.
        """
        try:
            term = self.terms[bar_id]
            if above <= term <= below:
                return (bar_id, term)
            else:
                # no term found with requested properties
                return (None, -np.inf)
        except KeyError:
            # bar ID does not exist
            return (None, -np.inf)

    def min_bif(self, bif_above=0.0, bif_below=np.inf):
        """Returns the id and value of the minimum bifurcation whose value is in threshold range.

        If no value is valid, returns infinity (np.inf) and therefore the index is None.
        """
        if np.isinf(bif_above):
            bif_above = 0.0
        for bifurcation in self.bifs.items():
            bifurcation_value = bifurcation[1]
            if bif_below >= bifurcation_value >= bif_above:
                return bifurcation
        return (None, np.inf)

    def min_term(self, term_above=0.0, term_below=np.inf):
        """Returns the id and value of the minimum termination whose value is in threshold range.

        If no value is valid, returns zero, the section will terminate and therefore the index is
        None.
        """
        if np.isinf(term_above):
            term_above = 0.0
        for termination in self.terms.items():
            termination_value = termination[1]
            if term_below >= termination_value >= term_above:
                return termination
        return (None, 0)

    def max_term(self):
        """Returns the id and value of the maximum termination.

        Termination list cannot be empty. This means the growth should have stopped, and therefore
        it will results in a 'StopIteration' error
        """
        return next(reversed(self.terms.items()))

    def curate_stop_criterion(self, parent_stop, child_stop):
        """Checks if the children stop criterion is compatible with parent.

        The child bar's length should be smaller than the current bar's length.
        This process ensures that each branch can only generate smaller branches.
        The criteria to ensure this statement is `True` are the following:

        * parent_stop.ref <= child_stop.bif <= parent_stop.term or child_stop.bif = inf
        * child_stop.term <= parent_stop.term
        * term(child_stop.bif) <= parent_stop.term

        Args:
            parent_stop (TMDStop): stop criteria of parent section.
            target_stop (TMDStop): proposed stop criteria for child.

        Returns:
            TMDStop: The next bar for which the expected length of the child branch is smaller than
            current one for both bif and term.
        """
        MAX_ref = parent_stop.term
        target_stop = copy.deepcopy(child_stop)

        # Case 0. Incompatibility checks
        # One of the assumed conditions is wrong
        # This should not happen, therefore growth should stop!
        if target_stop.term > MAX_ref:
            raise NeuroTSError("broken pipeline")
        if (not np.isinf(target_stop.bif)) and (target_stop.bif > MAX_ref):
            raise NeuroTSError("broken pipeline")

        # Case 1. Requirements fulfilled for inputs ref, bif, term
        if self.get_term(target_stop.bif_id) <= target_stop.term:
            return target_stop

        # Case 2. A new bifurcation needs to be identified
        bif_id, bif = self.select_compatible_bif(
            below_bif=parent_stop.ref,
            above_bif=MAX_ref,
            below_term=parent_stop.ref,
            above_term=target_stop.term,
        )
        target_stop.update_bif(bif_id, bif)
        return target_stop

    def select_compatible_bif(self, below_bif, above_bif, below_term, above_term):
        """Finds a bifurcation within the barcode.

        The bifurcation must fulfil both bif and term criteria / boundaries:
        below_bif <= bif <= above_bif
        below_term <= term <= above_term
        """
        # Search bar according to minimum bifurcation
        for bif_id, bif in self.bifs.items():
            corresp_term = self.get_term(bif_id)
            if (below_bif <= bif <= above_bif) and (below_term <= corresp_term <= above_term):
                # Define new termination corresponding to bifurcation
                return (bif_id, bif)
        return (None, np.inf)



"""Test neurots.astrocyte.section code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import itertools

import numpy as np
from mock import Mock
from numpy import testing as npt

from neurots.astrocyte.section import SectionSpatialGrower
from neurots.generate.tree import SectionParameters

POINT_CLOUD_POINTS = np.array(
    [
        [0.0, 0.0, 0.0],
        [0.1, 0.1, 0.1],
        [0.2, 0.2, 0.2],
        [0.3, 0.3, 0.3],
        [0.5, 0.5, 0.5],
        [0.6, 0.6, 0.6],
    ]
)

SECTION_PARAMETERS = SectionParameters(randomness=0.5, targeting=0.5, scale_prob=1.0, history=0.0)


def _point_cloud():

    point_cloud = Mock(points=POINT_CLOUD_POINTS, remove_hemisphere=Mock())

    values = [
        np.array([0.0, 0.0, 1.0]),
        np.array([0.0, 1.0, 0.0]),
        np.array([1.0, 0.0, 0.0]),
    ]

    point_cloud.nearest_neighbor_direction = Mock(side_effect=itertools.cycle(values))

    return point_cloud


class MockMorphologyPoints:
    def __init__(self):
        self._points = np.array([[0.0, 1.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])

    def append(self, point):
        self._points = np.vstack((self._points, point))

    @property
    def data(self):
        return self._points


def _spatial_context():
    return Mock(
        point_cloud=_point_cloud(),
        collision_handle=Mock(return_value=False),
        morphology_points=MockMorphologyPoints(),
    )


def _create_section_spatial_grower(process_type):

    start_point = np.array([0.0, 1.0, 2.0])
    direction = np.array([0.0, 1.0, 0.0])

    parent = Mock()

    step_size_distribution = Mock()
    # draw positive will cycle between 0.1, and 0.2
    step_size_distribution.draw_positive = Mock(side_effect=itertools.cycle([0.1, 0.2]))
    step_size_distribution.params = {"mean": 0.1, "std": 0.001}

    return SectionSpatialGrower(
        parent=parent,
        children=[],
        first_point=start_point,
        direction=direction,
        parameters=SECTION_PARAMETERS,
        process=process_type,
        stop_criteria={"TMD": Mock(bif=0.0, term=0.0)},
        step_size_distribution=step_size_distribution,
        pathlength=5.2,
        context=_spatial_context(),
    )


def test_section_spatial_grower__construction():

    grower = _create_section_spatial_grower("major")

    assert grower.process == "major"

    npt.assert_allclose(grower.pathlength, 5.2)
    npt.assert_allclose(grower.params, SECTION_PARAMETERS)
    npt.assert_allclose(grower.direction, [0.0, 1.0, 0.0])

    npt.assert_equal(len(grower.morphology_points.data), 3)

    npt.assert_allclose(grower.history(), [0.0, 0.0, 0.0])
    npt.assert_allclose(grower.last_point, [0.0, 1.0, 2.0])


def test_section_spatial_grower__first_point():

    grower = _create_section_spatial_grower("major")
    grower.first_point()

    expected_point = [0.0, 1.1, 2.0]

    npt.assert_allclose(grower.direction, [0.0, 1.0, 0.0])
    npt.assert_allclose(grower.latest_directions, [[0.0, 1.0, 0.0]])
    npt.assert_allclose(grower.points, [[0.0, 1.0, 2.0], expected_point])

    npt.assert_allclose(
        grower.morphology_points.data,
        [[0.0, 1.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], expected_point],
    )


def test_section_spatial_grower__next_direction():

    grower = _create_section_spatial_grower("major")

    grower.first_point()

    current_point = grower.last_point

    npt.assert_allclose(grower._neighbor_contribution(current_point), [0.0, 0.0, 1.0])

    npt.assert_allclose(grower.history(), [0.0, 1.0, 0.0])
    npt.assert_allclose(grower.next_direction(current_point), [0.0, 1.0, 0.0])


def test_section_spatial_grower__next_point():

    grower = _create_section_spatial_grower("major")
    grower.first_point()
    grower.next()

    expected_next_direction = 0.5 * grower.direction + 0.5 * np.array([0.0, 0.0, 1.0])
    expected_next_direction /= np.linalg.norm(expected_next_direction)

    expected_next_point = np.array([0.0, 1.1, 2.0]) + 0.2 * expected_next_direction

    npt.assert_allclose(grower.latest_directions, [[0.0, 1.0, 0.0], expected_next_direction])

    npt.assert_allclose(grower.points, [[0.0, 1.0, 2.0], [0.0, 1.1, 2.0], expected_next_point])

    npt.assert_allclose(
        grower.morphology_points.data,
        [
            [0.0, 1.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 1.1, 2.0],
            expected_next_point,
        ],
    )

    npt.assert_allclose(grower.pathlength, 0.2 + 0.1 + 5.2)



"""Definition of distributions to sample from."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


class Distr:
    """Class of custom distributions.

    Args:
        params (dict): The parameters of the distribution.
        random_generator (numpy.random.Generator): The random number generator to use.
    """

    def __init__(self, params, random_generator=np.random):
        """Object of statistical distribution."""
        self.type, distr_params = next(iter(params.items()))
        self._rng = random_generator
        self.loc = 0.0
        self.scale = 1.0
        self.set_distribution(distr_params)

    @staticmethod
    def norm(params):
        """Return loc, scale as expected from scipy from mean, std data."""
        return "standard_normal", params["mean"], params["std"]

    @staticmethod
    def uniform(params):
        """Return loc, scale as expected from scipy from min, max of a uniform."""
        return "uniform", params["min"], params["max"] - params["min"]

    @staticmethod
    def expon(params):
        """Return loc, scale as expected from scipy from mean, std data."""
        return "standard_exponential", params["loc"], 1.0 / params["lambda"]

    def set_distribution(self, params):
        """Return a statistical distribution according to input parameters."""
        # If distribution is a statistical distribution
        if self.type != "data":
            name, self.loc, self.scale = getattr(self, self.type)(params)
            self.distribution = getattr(self._rng, name)
        # If distribution consists of data we reformat input
        else:
            w = np.array(params["weights"], dtype=float)
            b = np.array(params["bins"])
            self.distribution = {"bins": b, "weights": w / np.sum(w)}

    def draw(self):
        """Return a sampled number."""
        if self.type == "data":
            return self._rng.choice(self.distribution["bins"], p=self.distribution["weights"])

        return self.loc + self.scale * self.distribution()

    def draw_positive(self):
        """Return a positive sampled number."""
        if self.type == "data":
            positives = np.where(self.distribution["bins"] > 0)
            return self._rng.choice(
                self.distribution["bins"][positives],
                p=self.distribution["weights"][positives],
            )

        if self.scale == 0:
            if self.loc >= 0:
                return self.loc
            else:
                raise ValueError(
                    "The 'loc' of the distribution must be >= 0 when 'scale' == 0 (loc == "
                    f"{self.scale})"
                )

        val = self.loc + self.scale * self.distribution()
        while val <= 0:
            val = self.loc + self.scale * self.distribution()
        return val


def d_transform(distr, funct, **kwargs):
    """Transform a distribution according to a selected function."""
    transf = {}
    for k in distr.keys():
        transf[k] = {}
        for j in distr[k].keys():
            transf[k][j] = funct(distr[k][j], **kwargs)

    return transf


def soma_size(distrib, random_generator=np.random):
    """Return a random soma radius as sampled from a distribution plus some constraints."""
    soma_d = Distr(distrib["soma"]["size"], random_generator)
    return soma_d.draw_positive()


def n_neurites(distrib, random_generator=np.random):
    """Return a number of neurites as sampled from a distribution plus some constraints.

    It ensures the number will be an INT.
    """
    neurites_d = Distr(distrib, random_generator)
    numtrees = int(neurites_d.draw())
    return numtrees


def trunk_angles(distrib, N, random_generator=np.random):
    """Return N relative angles, depending on the input distribution."""
    trunks_d = Distr(distrib["trunk"]["orientation_deviation"], random_generator)
    angles = [trunks_d.draw() for _ in range(N - 1)]
    angles = angles + [sum(angles)]
    return angles


def trunk_absolute_angles(distrib, N, random_generator=np.random):
    """Return N absolute angles, depending on the input distribution."""
    trunks_d = Distr(distrib["trunk"]["absolute_elevation_deviation"], random_generator)
    return [trunks_d.draw() for _ in range(N)]


def azimuth_angles(distrib, N, random_generator=np.random):
    """Return N azimuth angles, depending on the input distribution."""
    trunks_d = Distr(d_transform(distrib["trunk"]["azimuth"], np.cos), random_generator)
    tmp = np.array([trunks_d.draw() for _ in range(N)])
    angles = np.arccos(tmp)
    return angles


def ph(phs, random_generator=np.random):
    """Samples randomly a persistence diagram from the input distribution."""
    index = random_generator.choice(len(phs))
    return phs[index]



"""NeuroTS class: Tree."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import json
import logging
from collections import namedtuple

import numpy as np
from morphio import PointLevel
from morphio import SectionType

from neurots.generate.algorithms import basicgrower
from neurots.generate.algorithms import tmdgrower
from neurots.generate.section import SectionGrower
from neurots.generate.section import SectionGrowerPath
from neurots.generate.section import SectionGrowerTMD
from neurots.morphmath import sample
from neurots.utils import NeuroTSError

L = logging.getLogger("neurots")

# LAMBDA: parameter that defines the slope of exponential probability
LAMBDA = 1.0

growth_algorithms = {
    "tmd": tmdgrower.TMDAlgo,
    "tmd_apical": tmdgrower.TMDApicalAlgo,
    "tmd_gradient": tmdgrower.TMDGradientAlgo,
    "axon_trunk": basicgrower.AxonAlgo,
    "trunk": basicgrower.TrunkAlgo,
}

section_growers = {
    "radial_distances": SectionGrowerTMD,
    "path_distances": SectionGrowerPath,
    "trunk_length": SectionGrower,
}


# Section grower parameters
SectionParameters = namedtuple(
    "SectionParameters", ["randomness", "targeting", "scale_prob", "history"]
)


def _create_section_parameters(input_dict):
    """Create section parameters from input dictionary.

    Args:
        input_dict (dict): Input dictionary with ``randomness`` and ``targeting`` entries.

    Returns:
        SectionParameters: The section parameters.
    """
    randomness = np.clip(input_dict["randomness"], 0.0, 1.0)
    targeting = np.clip(input_dict["targeting"], 0.0, 1.0)
    history = np.clip(1.0 - randomness - targeting, 0.0, 1.0)

    parameters = SectionParameters(
        randomness=randomness, targeting=targeting, scale_prob=LAMBDA, history=history
    )

    try:
        assert np.isclose(randomness + targeting + history, 1.0)
        L.debug("Section Parameters: %s", parameters)
        return parameters

    except AssertionError as err:
        msg = f"Parameters randomness, targeting and history do not sum to 1:\n{parameters}"
        L.error(msg)
        raise NeuroTSError(msg) from err


class TreeGrower:
    """Tree class.

    Args:
        neuron (morphio.mut.Morphology): The morphology in which groups and points are stored.
        initial_direction (list[float]): 3D vector that defines the starting direction of the tree.
        initial_point (list[float]): 3D vector that defines the starting point of the tree.
        parameters (dict): A dictionary with ``tree_type``, ``radius``, ``randomness`` and
            ``targeting`` keys.
        distributions (dict): The distributions used.
        context (Any): The context used for the tree.
        random_generator (numpy.random.Generator): The random number generator to use.
    """

    def __init__(
        self,
        neuron,
        initial_direction,
        initial_point,
        parameters,
        distributions,
        skip_validation=False,
        context=None,
        random_generator=np.random,
    ):
        """Constructor of TreeGrower object."""
        self.neuron = neuron
        self.direction = initial_direction
        self.point = initial_point
        self.type = parameters["tree_type"]  # 2: axon, 3: basal, 4: apical, 5: other
        self.params = parameters
        self.distr = distributions
        self.active_sections = []
        self.context = context
        self.skip_validation = skip_validation
        self._rng = random_generator

        # Creates the distribution from which the segment lengths
        # To sample a new seg_len call self.seg_len.draw()
        self.seg_length_distr = sample.Distr(self.params["step_size"], random_generator=self._rng)
        self._section_parameters = _create_section_parameters(parameters)
        self.growth_algo = self._initialize_algorithm()

    def _initialize_algorithm(self):
        """Initialization steps for TreeGrower."""
        grow_meth = growth_algorithms[self.params["growth_method"]]

        growth_algo = grow_meth(
            input_data=self.distr,
            params=self.params,
            start_point=self.point,
            skip_validation=self.skip_validation,
            context=self.context,
            random_generator=self._rng,
        )

        stop, num_sec = growth_algo.initialize()

        sec = self.add_section(
            parent=None,
            direction=self.direction,
            first_point=list(self.point),
            stop=copy.deepcopy(stop),
            process="major",
            pathlength=0.0,
            children=2 if num_sec > 1 else 0,
        )
        # First section of the tree has at least one point.
        sec.first_point()

        return growth_algo

    def add_section(
        self, parent, direction, first_point, stop, pathlength, process=None, children=0
    ):
        """Generates a section from the parent section "act" from all the required information.

        The section is added to the neuron.sections and activated.

        Args:
            parent (morphio.Section): The parent of the section.
            direction (list[float]): The direction of the section.
            first_point (list[float]): The first point of the section.
            stop (dict):The stop criteria used for this section.
            pathlength (float): The path length of the section.
            process (str): The process type.
            children (int): The number of children.
        """
        SGrower = section_growers[self.params["metric"]]

        sec_grower = SGrower(
            parent=parent,
            first_point=first_point,
            direction=direction,
            parameters=self._section_parameters,
            children=children,
            process=process,
            stop_criteria=copy.deepcopy(stop),
            step_size_distribution=self.seg_length_distr,
            pathlength=pathlength,
            context=self.context,
            random_generator=self._rng,
        )

        self.active_sections.append(sec_grower)
        return sec_grower

    def end(self):
        """Ends the growth."""
        return not bool(self.active_sections)

    @staticmethod
    def order_per_process(secs):
        """Orders sections according to process type, major first."""
        return np.copy(secs)[np.argsort([ss.process for ss in secs])]

    @staticmethod
    def order_per_bif(secs):
        """Orders sections according to bifurcation times."""
        ordered_list = np.argsort([ss.stop_criteria["TMD"].bif for ss in secs])
        return np.copy(secs)[ordered_list]

    def append_section(self, section):
        """Append section to the MorphIO neuron.

        Args:
            section (SectionGrowerPath): The section that is going to be appended.

        Returns:
            section (morphio.Section): The new appended section.
        """
        if section.parent:
            append_fun = section.parent.append_section
        else:
            append_fun = self.neuron.append_root_section

        if L.level == logging.DEBUG:  # pragma: no cover
            data = {
                "parent": section.parent.id if section.parent else None,
                "coord": np.vstack(section.points).tolist(),
                "radius": [self.params["radius"] * 2] * len(section.points),
                "type": int(SectionType(self.params["tree_type"])),
            }
            L.debug("appended_data=%s", json.dumps(data))

        return append_fun(
            PointLevel(
                np.array(section.points).tolist(),
                [self.params["radius"] * 2] * len(section.points),
            ),
            SectionType(self.params["tree_type"]),
        )

    def next_point(self):
        """Operates the tree growth according to the selected algorithm."""
        if not isinstance(self.growth_algo, basicgrower.TrunkAlgo):
            ordered_sections = self.order_per_bif(self.active_sections)
        else:
            # TrunkAlgo does not keep track of the bifurcations so it is not
            # possible to order per bifurcation
            ordered_sections = np.copy(self.active_sections)

        for section_grower in ordered_sections:
            # the current section_grower is generated
            # In here the stop criterion can be modified accordingly
            state = self.growth_algo.extend(section_grower)

            if state != "continue":
                section = self.append_section(section_grower)

                if state == "bifurcate":
                    # Save the final normed direction of parent
                    latest = section_grower.latest_directions[-1]
                    section_grower.id = section.id
                    # the current section_grower bifurcates
                    # Returns two section_grower dictionaries: (S1, S2)
                    for child_section in self.growth_algo.bifurcate(section_grower):
                        child = self.add_section(
                            parent=section, pathlength=section_grower.pathlength, **child_section
                        )
                        # Copy the final normed direction of parent to all children
                        child.latest_directions.append(latest)
                        # Generate the first point of the section
                        child.first_point()
                    self.active_sections.remove(section_grower)

                elif state == "terminate":
                    # the current section_grower terminates
                    self.growth_algo.terminate(section_grower)
                    self.active_sections.remove(section_grower)

                # TODO: Can the state be something else than "bifurcate" or "terminate" here?
                # If not we could add an else statement to raise a warning or an exception if the
                # value is anything else.



"""Test neurots.morphmath.rotation code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
from numpy.testing import assert_array_almost_equal

import neurots.morphmath.rotation as test_module


def test_vector_from_spherical():
    assert_array_almost_equal(
        test_module.vector_from_spherical(1, 1), [0.454649, 0.708073, 0.540302]
    )


def test_rotate_vector():
    assert_array_almost_equal(
        test_module.rotate_vector([1, 1, 1], [2, 3, 4], 2),
        [0.293989, 1.240039, 1.172976],
    )


def test_angle3D():
    assert_array_almost_equal(test_module.angle3D([1, 1, 1], [2, 3, 4]), 0.265729)
